# pylint: disable=too-many-lines,too-many-statements
# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------
import datetime
from io import IOBase
from typing import Any, AsyncIterable, Callable, Dict, IO, Optional, TypeVar, Union, overload
import urllib.parse

from azure.core.async_paging import AsyncItemPaged, AsyncList
from azure.core.exceptions import (
    ClientAuthenticationError,
    HttpResponseError,
    ResourceExistsError,
    ResourceNotFoundError,
    ResourceNotModifiedError,
    map_error,
)
from azure.core.pipeline import PipelineResponse
from azure.core.pipeline.transport import AsyncHttpResponse
from azure.core.rest import HttpRequest
from azure.core.tracing.decorator import distributed_trace
from azure.core.tracing.decorator_async import distributed_trace_async
from azure.core.utils import case_insensitive_dict

from ... import models as _models
from ..._vendor import _convert_request
from ...operations._chat_thread_operations import (
    build_add_chat_participants_request,
    build_delete_chat_message_request,
    build_get_chat_message_request,
    build_get_chat_thread_properties_request,
    build_list_chat_messages_request,
    build_list_chat_participants_request,
    build_list_chat_read_receipts_request,
    build_remove_chat_participant_request,
    build_send_chat_message_request,
    build_send_chat_read_receipt_request,
    build_send_typing_notification_request,
    build_update_chat_message_request,
    build_update_chat_thread_properties_request,
)

T = TypeVar("T")
ClsType = Optional[Callable[[PipelineResponse[HttpRequest, AsyncHttpResponse], T, Dict[str, Any]], Any]]


class ChatThreadOperations:
    """
    .. warning::
        **DO NOT** instantiate this class directly.

        Instead, you should access the following operations through
        :class:`~azure.communication.chat.aio.AzureCommunicationChatService`'s
        :attr:`chat_thread` attribute.
    """

    models = _models

    def __init__(self, *args, **kwargs) -> None:
        input_args = list(args)
        self._client = input_args.pop(0) if input_args else kwargs.pop("client")
        self._config = input_args.pop(0) if input_args else kwargs.pop("config")
        self._serialize = input_args.pop(0) if input_args else kwargs.pop("serializer")
        self._deserialize = input_args.pop(0) if input_args else kwargs.pop("deserializer")

    @distributed_trace
    def list_chat_read_receipts(
        self, chat_thread_id: str, max_page_size: Optional[int] = None, skip: Optional[int] = None, **kwargs: Any
    ) -> AsyncIterable["_models.ChatMessageReadReceipt"]:
        """Gets chat message read receipts for a thread.

        Gets chat message read receipts for a thread.

        :param chat_thread_id: Thread id to get the chat message read receipts for. Required.
        :type chat_thread_id: str
        :param max_page_size: The maximum number of chat message read receipts to be returned per page.
         Default value is None.
        :type max_page_size: int
        :param skip: Skips chat message read receipts up to a specified position in response. Default
         value is None.
        :type skip: int
        :return: An iterator like instance of either ChatMessageReadReceipt or the result of
         cls(response)
        :rtype:
         ~azure.core.async_paging.AsyncItemPaged[~azure.communication.chat.models.ChatMessageReadReceipt]
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        _headers = kwargs.pop("headers", {}) or {}
        _params = case_insensitive_dict(kwargs.pop("params", {}) or {})

        api_version: str = kwargs.pop("api_version", _params.pop("api-version", self._config.api_version))
        cls: ClsType[_models.ChatMessageReadReceiptsCollection] = kwargs.pop("cls", None)

        error_map = {
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
            401: lambda response: ClientAuthenticationError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            403: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            429: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            503: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        def prepare_request(next_link=None):
            if not next_link:

                _request = build_list_chat_read_receipts_request(
                    chat_thread_id=chat_thread_id,
                    max_page_size=max_page_size,
                    skip=skip,
                    api_version=api_version,
                    headers=_headers,
                    params=_params,
                )
                _request = _convert_request(_request)
                path_format_arguments = {
                    "endpoint": self._serialize.url(
                        "self._config.endpoint", self._config.endpoint, "str", skip_quote=True
                    ),
                }
                _request.url = self._client.format_url(_request.url, **path_format_arguments)

            else:
                # make call to next link with the client's api-version
                _parsed_next_link = urllib.parse.urlparse(next_link)
                _next_request_params = case_insensitive_dict(
                    {
                        key: [urllib.parse.quote(v) for v in value]
                        for key, value in urllib.parse.parse_qs(_parsed_next_link.query).items()
                    }
                )
                _next_request_params["api-version"] = self._config.api_version
                _request = HttpRequest(
                    "GET", urllib.parse.urljoin(next_link, _parsed_next_link.path), params=_next_request_params
                )
                _request = _convert_request(_request)
                path_format_arguments = {
                    "endpoint": self._serialize.url(
                        "self._config.endpoint", self._config.endpoint, "str", skip_quote=True
                    ),
                }
                _request.url = self._client.format_url(_request.url, **path_format_arguments)
                _request.method = "GET"
            return _request

        async def extract_data(pipeline_response):
            deserialized = self._deserialize("ChatMessageReadReceiptsCollection", pipeline_response)
            list_of_elem = deserialized.value
            if cls:
                list_of_elem = cls(list_of_elem)  # type: ignore
            return deserialized.next_link or None, AsyncList(list_of_elem)

        async def get_next(next_link=None):
            _request = prepare_request(next_link)

            _stream = False
            pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
                _request, stream=_stream, **kwargs
            )
            response = pipeline_response.http_response

            if response.status_code not in [200]:
                map_error(status_code=response.status_code, response=response, error_map=error_map)  # type: ignore
                raise HttpResponseError(response=response)

            return pipeline_response

        return AsyncItemPaged(get_next, extract_data)

    @overload
    async def send_chat_read_receipt(  # pylint: disable=inconsistent-return-statements
        self,
        chat_thread_id: str,
        send_read_receipt_request: _models.SendReadReceiptRequest,
        *,
        content_type: str = "application/json",
        **kwargs: Any
    ) -> None:
        """Sends a read receipt event to a thread, on behalf of a user.

        Sends a read receipt event to a thread, on behalf of a user.

        :param chat_thread_id: Thread id to send the read receipt event to. Required.
        :type chat_thread_id: str
        :param send_read_receipt_request: Read receipt details. Required.
        :type send_read_receipt_request: ~azure.communication.chat.models.SendReadReceiptRequest
        :keyword content_type: Body Parameter content-type. Content type parameter for JSON body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: None or the result of cls(response)
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """

    @overload
    async def send_chat_read_receipt(  # pylint: disable=inconsistent-return-statements
        self,
        chat_thread_id: str,
        send_read_receipt_request: IO[bytes],
        *,
        content_type: str = "application/json",
        **kwargs: Any
    ) -> None:
        """Sends a read receipt event to a thread, on behalf of a user.

        Sends a read receipt event to a thread, on behalf of a user.

        :param chat_thread_id: Thread id to send the read receipt event to. Required.
        :type chat_thread_id: str
        :param send_read_receipt_request: Read receipt details. Required.
        :type send_read_receipt_request: IO[bytes]
        :keyword content_type: Body Parameter content-type. Content type parameter for binary body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: None or the result of cls(response)
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """

    @distributed_trace_async
    async def send_chat_read_receipt(  # pylint: disable=inconsistent-return-statements
        self,
        chat_thread_id: str,
        send_read_receipt_request: Union[_models.SendReadReceiptRequest, IO[bytes]],
        **kwargs: Any
    ) -> None:
        """Sends a read receipt event to a thread, on behalf of a user.

        Sends a read receipt event to a thread, on behalf of a user.

        :param chat_thread_id: Thread id to send the read receipt event to. Required.
        :type chat_thread_id: str
        :param send_read_receipt_request: Read receipt details. Is either a SendReadReceiptRequest type
         or a IO[bytes] type. Required.
        :type send_read_receipt_request: ~azure.communication.chat.models.SendReadReceiptRequest or
         IO[bytes]
        :return: None or the result of cls(response)
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        error_map = {
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
            401: lambda response: ClientAuthenticationError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            403: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            429: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            503: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = case_insensitive_dict(kwargs.pop("headers", {}) or {})
        _params = case_insensitive_dict(kwargs.pop("params", {}) or {})

        api_version: str = kwargs.pop("api_version", _params.pop("api-version", self._config.api_version))
        content_type: Optional[str] = kwargs.pop("content_type", _headers.pop("Content-Type", None))
        cls: ClsType[None] = kwargs.pop("cls", None)

        content_type = content_type or "application/json"
        _json = None
        _content = None
        if isinstance(send_read_receipt_request, (IOBase, bytes)):
            _content = send_read_receipt_request
        else:
            _json = self._serialize.body(send_read_receipt_request, "SendReadReceiptRequest")

        _request = build_send_chat_read_receipt_request(
            chat_thread_id=chat_thread_id,
            api_version=api_version,
            content_type=content_type,
            json=_json,
            content=_content,
            headers=_headers,
            params=_params,
        )
        _request = _convert_request(_request)
        path_format_arguments = {
            "endpoint": self._serialize.url("self._config.endpoint", self._config.endpoint, "str", skip_quote=True),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)  # type: ignore
            raise HttpResponseError(response=response)

        if cls:
            return cls(pipeline_response, None, {})  # type: ignore

    @overload
    async def send_chat_message(
        self,
        chat_thread_id: str,
        send_chat_message_request: _models.SendChatMessageRequest,
        *,
        content_type: str = "application/json",
        **kwargs: Any
    ) -> _models.SendChatMessageResult:
        """Sends a message to a thread.

        Sends a message to a thread.

        :param chat_thread_id: The thread id to send the message to. Required.
        :type chat_thread_id: str
        :param send_chat_message_request: Details of the message to send. Required.
        :type send_chat_message_request: ~azure.communication.chat.models.SendChatMessageRequest
        :keyword content_type: Body Parameter content-type. Content type parameter for JSON body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: SendChatMessageResult or the result of cls(response)
        :rtype: ~azure.communication.chat.models.SendChatMessageResult
        :raises ~azure.core.exceptions.HttpResponseError:
        """

    @overload
    async def send_chat_message(
        self,
        chat_thread_id: str,
        send_chat_message_request: IO[bytes],
        *,
        content_type: str = "application/json",
        **kwargs: Any
    ) -> _models.SendChatMessageResult:
        """Sends a message to a thread.

        Sends a message to a thread.

        :param chat_thread_id: The thread id to send the message to. Required.
        :type chat_thread_id: str
        :param send_chat_message_request: Details of the message to send. Required.
        :type send_chat_message_request: IO[bytes]
        :keyword content_type: Body Parameter content-type. Content type parameter for binary body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: SendChatMessageResult or the result of cls(response)
        :rtype: ~azure.communication.chat.models.SendChatMessageResult
        :raises ~azure.core.exceptions.HttpResponseError:
        """

    @distributed_trace_async
    async def send_chat_message(
        self,
        chat_thread_id: str,
        send_chat_message_request: Union[_models.SendChatMessageRequest, IO[bytes]],
        **kwargs: Any
    ) -> _models.SendChatMessageResult:
        """Sends a message to a thread.

        Sends a message to a thread.

        :param chat_thread_id: The thread id to send the message to. Required.
        :type chat_thread_id: str
        :param send_chat_message_request: Details of the message to send. Is either a
         SendChatMessageRequest type or a IO[bytes] type. Required.
        :type send_chat_message_request: ~azure.communication.chat.models.SendChatMessageRequest or
         IO[bytes]
        :return: SendChatMessageResult or the result of cls(response)
        :rtype: ~azure.communication.chat.models.SendChatMessageResult
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        error_map = {
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
            401: lambda response: ClientAuthenticationError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            403: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            429: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            503: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = case_insensitive_dict(kwargs.pop("headers", {}) or {})
        _params = case_insensitive_dict(kwargs.pop("params", {}) or {})

        api_version: str = kwargs.pop("api_version", _params.pop("api-version", self._config.api_version))
        content_type: Optional[str] = kwargs.pop("content_type", _headers.pop("Content-Type", None))
        cls: ClsType[_models.SendChatMessageResult] = kwargs.pop("cls", None)

        content_type = content_type or "application/json"
        _json = None
        _content = None
        if isinstance(send_chat_message_request, (IOBase, bytes)):
            _content = send_chat_message_request
        else:
            _json = self._serialize.body(send_chat_message_request, "SendChatMessageRequest")

        _request = build_send_chat_message_request(
            chat_thread_id=chat_thread_id,
            api_version=api_version,
            content_type=content_type,
            json=_json,
            content=_content,
            headers=_headers,
            params=_params,
        )
        _request = _convert_request(_request)
        path_format_arguments = {
            "endpoint": self._serialize.url("self._config.endpoint", self._config.endpoint, "str", skip_quote=True),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [201]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)  # type: ignore
            raise HttpResponseError(response=response)

        deserialized = self._deserialize("SendChatMessageResult", pipeline_response)

        if cls:
            return cls(pipeline_response, deserialized, {})  # type: ignore

        return deserialized  # type: ignore

    @distributed_trace
    def list_chat_messages(
        self,
        chat_thread_id: str,
        max_page_size: Optional[int] = None,
        start_time: Optional[datetime.datetime] = None,
        **kwargs: Any
    ) -> AsyncIterable["_models.ChatMessage"]:
        """Gets a list of messages from a thread.

        Gets a list of messages from a thread.

        :param chat_thread_id: The thread id of the message. Required.
        :type chat_thread_id: str
        :param max_page_size: The maximum number of messages to be returned per page. Default value is
         None.
        :type max_page_size: int
        :param start_time: The earliest point in time to get messages up to. The timestamp should be in
         RFC3339 format: ``yyyy-MM-ddTHH:mm:ssZ``. Default value is None.
        :type start_time: ~datetime.datetime
        :return: An iterator like instance of either ChatMessage or the result of cls(response)
        :rtype: ~azure.core.async_paging.AsyncItemPaged[~azure.communication.chat.models.ChatMessage]
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        _headers = kwargs.pop("headers", {}) or {}
        _params = case_insensitive_dict(kwargs.pop("params", {}) or {})

        api_version: str = kwargs.pop("api_version", _params.pop("api-version", self._config.api_version))
        cls: ClsType[_models.ChatMessagesCollection] = kwargs.pop("cls", None)

        error_map = {
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
            401: lambda response: ClientAuthenticationError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            403: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            429: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            503: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        def prepare_request(next_link=None):
            if not next_link:

                _request = build_list_chat_messages_request(
                    chat_thread_id=chat_thread_id,
                    max_page_size=max_page_size,
                    start_time=start_time,
                    api_version=api_version,
                    headers=_headers,
                    params=_params,
                )
                _request = _convert_request(_request)
                path_format_arguments = {
                    "endpoint": self._serialize.url(
                        "self._config.endpoint", self._config.endpoint, "str", skip_quote=True
                    ),
                }
                _request.url = self._client.format_url(_request.url, **path_format_arguments)

            else:
                # make call to next link with the client's api-version
                _parsed_next_link = urllib.parse.urlparse(next_link)
                _next_request_params = case_insensitive_dict(
                    {
                        key: [urllib.parse.quote(v) for v in value]
                        for key, value in urllib.parse.parse_qs(_parsed_next_link.query).items()
                    }
                )
                _next_request_params["api-version"] = self._config.api_version
                _request = HttpRequest(
                    "GET", urllib.parse.urljoin(next_link, _parsed_next_link.path), params=_next_request_params
                )
                _request = _convert_request(_request)
                path_format_arguments = {
                    "endpoint": self._serialize.url(
                        "self._config.endpoint", self._config.endpoint, "str", skip_quote=True
                    ),
                }
                _request.url = self._client.format_url(_request.url, **path_format_arguments)
                _request.method = "GET"
            return _request

        async def extract_data(pipeline_response):
            deserialized = self._deserialize("ChatMessagesCollection", pipeline_response)
            list_of_elem = deserialized.value
            if cls:
                list_of_elem = cls(list_of_elem)  # type: ignore
            return deserialized.next_link or None, AsyncList(list_of_elem)

        async def get_next(next_link=None):
            _request = prepare_request(next_link)

            _stream = False
            pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
                _request, stream=_stream, **kwargs
            )
            response = pipeline_response.http_response

            if response.status_code not in [200]:
                map_error(status_code=response.status_code, response=response, error_map=error_map)  # type: ignore
                raise HttpResponseError(response=response)

            return pipeline_response

        return AsyncItemPaged(get_next, extract_data)

    @distributed_trace_async
    async def get_chat_message(self, chat_thread_id: str, chat_message_id: str, **kwargs: Any) -> _models.ChatMessage:
        """Gets a message by id.

        Gets a message by id.

        :param chat_thread_id: The thread id to which the message was sent. Required.
        :type chat_thread_id: str
        :param chat_message_id: The message id. Required.
        :type chat_message_id: str
        :return: ChatMessage or the result of cls(response)
        :rtype: ~azure.communication.chat.models.ChatMessage
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        error_map = {
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
            401: lambda response: ClientAuthenticationError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            403: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            429: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            503: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = kwargs.pop("headers", {}) or {}
        _params = case_insensitive_dict(kwargs.pop("params", {}) or {})

        api_version: str = kwargs.pop("api_version", _params.pop("api-version", self._config.api_version))
        cls: ClsType[_models.ChatMessage] = kwargs.pop("cls", None)

        _request = build_get_chat_message_request(
            chat_thread_id=chat_thread_id,
            chat_message_id=chat_message_id,
            api_version=api_version,
            headers=_headers,
            params=_params,
        )
        _request = _convert_request(_request)
        path_format_arguments = {
            "endpoint": self._serialize.url("self._config.endpoint", self._config.endpoint, "str", skip_quote=True),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)  # type: ignore
            raise HttpResponseError(response=response)

        deserialized = self._deserialize("ChatMessage", pipeline_response)

        if cls:
            return cls(pipeline_response, deserialized, {})  # type: ignore

        return deserialized  # type: ignore

    @overload
    async def update_chat_message(  # pylint: disable=inconsistent-return-statements
        self,
        chat_thread_id: str,
        chat_message_id: str,
        update_chat_message_request: _models.UpdateChatMessageRequest,
        *,
        content_type: str = "application/merge-patch+json",
        **kwargs: Any
    ) -> None:
        """Updates a message.

        Updates a message.

        :param chat_thread_id: The thread id to which the message was sent. Required.
        :type chat_thread_id: str
        :param chat_message_id: The message id. Required.
        :type chat_message_id: str
        :param update_chat_message_request: Details of the request to update the message. Required.
        :type update_chat_message_request: ~azure.communication.chat.models.UpdateChatMessageRequest
        :keyword content_type: Body Parameter content-type. Content type parameter for JSON body.
         Default value is "application/merge-patch+json".
        :paramtype content_type: str
        :return: None or the result of cls(response)
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """

    @overload
    async def update_chat_message(  # pylint: disable=inconsistent-return-statements
        self,
        chat_thread_id: str,
        chat_message_id: str,
        update_chat_message_request: IO[bytes],
        *,
        content_type: str = "application/merge-patch+json",
        **kwargs: Any
    ) -> None:
        """Updates a message.

        Updates a message.

        :param chat_thread_id: The thread id to which the message was sent. Required.
        :type chat_thread_id: str
        :param chat_message_id: The message id. Required.
        :type chat_message_id: str
        :param update_chat_message_request: Details of the request to update the message. Required.
        :type update_chat_message_request: IO[bytes]
        :keyword content_type: Body Parameter content-type. Content type parameter for binary body.
         Default value is "application/merge-patch+json".
        :paramtype content_type: str
        :return: None or the result of cls(response)
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """

    @distributed_trace_async
    async def update_chat_message(  # pylint: disable=inconsistent-return-statements
        self,
        chat_thread_id: str,
        chat_message_id: str,
        update_chat_message_request: Union[_models.UpdateChatMessageRequest, IO[bytes]],
        **kwargs: Any
    ) -> None:
        """Updates a message.

        Updates a message.

        :param chat_thread_id: The thread id to which the message was sent. Required.
        :type chat_thread_id: str
        :param chat_message_id: The message id. Required.
        :type chat_message_id: str
        :param update_chat_message_request: Details of the request to update the message. Is either a
         UpdateChatMessageRequest type or a IO[bytes] type. Required.
        :type update_chat_message_request: ~azure.communication.chat.models.UpdateChatMessageRequest or
         IO[bytes]
        :return: None or the result of cls(response)
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        error_map = {
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
            401: lambda response: ClientAuthenticationError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            403: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            429: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            503: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = case_insensitive_dict(kwargs.pop("headers", {}) or {})
        _params = case_insensitive_dict(kwargs.pop("params", {}) or {})

        api_version: str = kwargs.pop("api_version", _params.pop("api-version", self._config.api_version))
        content_type: Optional[str] = kwargs.pop("content_type", _headers.pop("Content-Type", None))
        cls: ClsType[None] = kwargs.pop("cls", None)

        content_type = content_type or "application/merge-patch+json"
        _json = None
        _content = None
        if isinstance(update_chat_message_request, (IOBase, bytes)):
            _content = update_chat_message_request
        else:
            _json = self._serialize.body(update_chat_message_request, "UpdateChatMessageRequest")

        _request = build_update_chat_message_request(
            chat_thread_id=chat_thread_id,
            chat_message_id=chat_message_id,
            api_version=api_version,
            content_type=content_type,
            json=_json,
            content=_content,
            headers=_headers,
            params=_params,
        )
        _request = _convert_request(_request)
        path_format_arguments = {
            "endpoint": self._serialize.url("self._config.endpoint", self._config.endpoint, "str", skip_quote=True),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [204]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)  # type: ignore
            raise HttpResponseError(response=response)

        if cls:
            return cls(pipeline_response, None, {})  # type: ignore

    @distributed_trace_async
    async def delete_chat_message(  # pylint: disable=inconsistent-return-statements
        self, chat_thread_id: str, chat_message_id: str, **kwargs: Any
    ) -> None:
        """Deletes a message.

        Deletes a message.

        :param chat_thread_id: The thread id to which the message was sent. Required.
        :type chat_thread_id: str
        :param chat_message_id: The message id. Required.
        :type chat_message_id: str
        :return: None or the result of cls(response)
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        error_map = {
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
            401: lambda response: ClientAuthenticationError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            403: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            429: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            503: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = kwargs.pop("headers", {}) or {}
        _params = case_insensitive_dict(kwargs.pop("params", {}) or {})

        api_version: str = kwargs.pop("api_version", _params.pop("api-version", self._config.api_version))
        cls: ClsType[None] = kwargs.pop("cls", None)

        _request = build_delete_chat_message_request(
            chat_thread_id=chat_thread_id,
            chat_message_id=chat_message_id,
            api_version=api_version,
            headers=_headers,
            params=_params,
        )
        _request = _convert_request(_request)
        path_format_arguments = {
            "endpoint": self._serialize.url("self._config.endpoint", self._config.endpoint, "str", skip_quote=True),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [204]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)  # type: ignore
            raise HttpResponseError(response=response)

        if cls:
            return cls(pipeline_response, None, {})  # type: ignore

    @distributed_trace
    def list_chat_participants(
        self, chat_thread_id: str, max_page_size: Optional[int] = None, skip: Optional[int] = None, **kwargs: Any
    ) -> AsyncIterable["_models.ChatParticipant"]:
        """Gets the participants of a thread.

        Gets the participants of a thread.

        :param chat_thread_id: Thread id to get participants for. Required.
        :type chat_thread_id: str
        :param max_page_size: The maximum number of participants to be returned per page. Default value
         is None.
        :type max_page_size: int
        :param skip: Skips participants up to a specified position in response. Default value is None.
        :type skip: int
        :return: An iterator like instance of either ChatParticipant or the result of cls(response)
        :rtype:
         ~azure.core.async_paging.AsyncItemPaged[~azure.communication.chat.models.ChatParticipant]
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        _headers = kwargs.pop("headers", {}) or {}
        _params = case_insensitive_dict(kwargs.pop("params", {}) or {})

        api_version: str = kwargs.pop("api_version", _params.pop("api-version", self._config.api_version))
        cls: ClsType[_models.ChatParticipantsCollection] = kwargs.pop("cls", None)

        error_map = {
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
            401: lambda response: ClientAuthenticationError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            403: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            429: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            503: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        def prepare_request(next_link=None):
            if not next_link:

                _request = build_list_chat_participants_request(
                    chat_thread_id=chat_thread_id,
                    max_page_size=max_page_size,
                    skip=skip,
                    api_version=api_version,
                    headers=_headers,
                    params=_params,
                )
                _request = _convert_request(_request)
                path_format_arguments = {
                    "endpoint": self._serialize.url(
                        "self._config.endpoint", self._config.endpoint, "str", skip_quote=True
                    ),
                }
                _request.url = self._client.format_url(_request.url, **path_format_arguments)

            else:
                # make call to next link with the client's api-version
                _parsed_next_link = urllib.parse.urlparse(next_link)
                _next_request_params = case_insensitive_dict(
                    {
                        key: [urllib.parse.quote(v) for v in value]
                        for key, value in urllib.parse.parse_qs(_parsed_next_link.query).items()
                    }
                )
                _next_request_params["api-version"] = self._config.api_version
                _request = HttpRequest(
                    "GET", urllib.parse.urljoin(next_link, _parsed_next_link.path), params=_next_request_params
                )
                _request = _convert_request(_request)
                path_format_arguments = {
                    "endpoint": self._serialize.url(
                        "self._config.endpoint", self._config.endpoint, "str", skip_quote=True
                    ),
                }
                _request.url = self._client.format_url(_request.url, **path_format_arguments)
                _request.method = "GET"
            return _request

        async def extract_data(pipeline_response):
            deserialized = self._deserialize("ChatParticipantsCollection", pipeline_response)
            list_of_elem = deserialized.value
            if cls:
                list_of_elem = cls(list_of_elem)  # type: ignore
            return deserialized.next_link or None, AsyncList(list_of_elem)

        async def get_next(next_link=None):
            _request = prepare_request(next_link)

            _stream = False
            pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
                _request, stream=_stream, **kwargs
            )
            response = pipeline_response.http_response

            if response.status_code not in [200]:
                map_error(status_code=response.status_code, response=response, error_map=error_map)  # type: ignore
                raise HttpResponseError(response=response)

            return pipeline_response

        return AsyncItemPaged(get_next, extract_data)

    @overload
    async def remove_chat_participant(  # pylint: disable=inconsistent-return-statements
        self,
        chat_thread_id: str,
        participant_communication_identifier: _models.CommunicationIdentifierModel,
        *,
        content_type: str = "application/json",
        **kwargs: Any
    ) -> None:
        """Remove a participant from a thread.

        Remove a participant from a thread.

        :param chat_thread_id: Thread id to remove the participant from. Required.
        :type chat_thread_id: str
        :param participant_communication_identifier: Id of the thread participant to remove from the
         thread. Required.
        :type participant_communication_identifier:
         ~azure.communication.chat.models.CommunicationIdentifierModel
        :keyword content_type: Body Parameter content-type. Content type parameter for JSON body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: None or the result of cls(response)
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """

    @overload
    async def remove_chat_participant(  # pylint: disable=inconsistent-return-statements
        self,
        chat_thread_id: str,
        participant_communication_identifier: IO[bytes],
        *,
        content_type: str = "application/json",
        **kwargs: Any
    ) -> None:
        """Remove a participant from a thread.

        Remove a participant from a thread.

        :param chat_thread_id: Thread id to remove the participant from. Required.
        :type chat_thread_id: str
        :param participant_communication_identifier: Id of the thread participant to remove from the
         thread. Required.
        :type participant_communication_identifier: IO[bytes]
        :keyword content_type: Body Parameter content-type. Content type parameter for binary body.
         Known values are: 'application/json', 'application/merge-patch+json'. Default value is
         "application/json".
        :paramtype content_type: str
        :return: None or the result of cls(response)
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """

    @distributed_trace_async
    async def remove_chat_participant(  # pylint: disable=inconsistent-return-statements
        self,
        chat_thread_id: str,
        participant_communication_identifier: Union[_models.CommunicationIdentifierModel, IO[bytes]],
        **kwargs: Any
    ) -> None:
        """Remove a participant from a thread.

        Remove a participant from a thread.

        :param chat_thread_id: Thread id to remove the participant from. Required.
        :type chat_thread_id: str
        :param participant_communication_identifier: Id of the thread participant to remove from the
         thread. Is either a CommunicationIdentifierModel type or a IO[bytes] type. Required.
        :type participant_communication_identifier:
         ~azure.communication.chat.models.CommunicationIdentifierModel or IO[bytes]
        :return: None or the result of cls(response)
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        error_map = {
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
            401: lambda response: ClientAuthenticationError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            403: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            429: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            503: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = case_insensitive_dict(kwargs.pop("headers", {}) or {})
        _params = case_insensitive_dict(kwargs.pop("params", {}) or {})

        api_version: str = kwargs.pop("api_version", _params.pop("api-version", self._config.api_version))
        content_type: Optional[str] = kwargs.pop("content_type", _headers.pop("Content-Type", None))
        cls: ClsType[None] = kwargs.pop("cls", None)

        content_type = content_type or "application/json"
        _json = None
        _content = None
        if isinstance(participant_communication_identifier, (IOBase, bytes)):
            _content = participant_communication_identifier
        else:
            _json = self._serialize.body(participant_communication_identifier, "CommunicationIdentifierModel")

        _request = build_remove_chat_participant_request(
            chat_thread_id=chat_thread_id,
            api_version=api_version,
            content_type=content_type,
            json=_json,
            content=_content,
            headers=_headers,
            params=_params,
        )
        _request = _convert_request(_request)
        path_format_arguments = {
            "endpoint": self._serialize.url("self._config.endpoint", self._config.endpoint, "str", skip_quote=True),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [204]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)  # type: ignore
            raise HttpResponseError(response=response)

        if cls:
            return cls(pipeline_response, None, {})  # type: ignore

    @overload
    async def add_chat_participants(
        self,
        chat_thread_id: str,
        add_chat_participants_request: _models.AddChatParticipantsRequest,
        *,
        content_type: str = "application/json",
        **kwargs: Any
    ) -> _models.AddChatParticipantsResult:
        """Adds thread participants to a thread. If participants already exist, no change occurs.

        Adds thread participants to a thread. If participants already exist, no change occurs.

        :param chat_thread_id: Id of the thread to add participants to. Required.
        :type chat_thread_id: str
        :param add_chat_participants_request: Thread participants to be added to the thread. Required.
        :type add_chat_participants_request:
         ~azure.communication.chat.models.AddChatParticipantsRequest
        :keyword content_type: Body Parameter content-type. Content type parameter for JSON body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: AddChatParticipantsResult or the result of cls(response)
        :rtype: ~azure.communication.chat.models.AddChatParticipantsResult
        :raises ~azure.core.exceptions.HttpResponseError:
        """

    @overload
    async def add_chat_participants(
        self,
        chat_thread_id: str,
        add_chat_participants_request: IO[bytes],
        *,
        content_type: str = "application/json",
        **kwargs: Any
    ) -> _models.AddChatParticipantsResult:
        """Adds thread participants to a thread. If participants already exist, no change occurs.

        Adds thread participants to a thread. If participants already exist, no change occurs.

        :param chat_thread_id: Id of the thread to add participants to. Required.
        :type chat_thread_id: str
        :param add_chat_participants_request: Thread participants to be added to the thread. Required.
        :type add_chat_participants_request: IO[bytes]
        :keyword content_type: Body Parameter content-type. Content type parameter for binary body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: AddChatParticipantsResult or the result of cls(response)
        :rtype: ~azure.communication.chat.models.AddChatParticipantsResult
        :raises ~azure.core.exceptions.HttpResponseError:
        """

    @distributed_trace_async
    async def add_chat_participants(
        self,
        chat_thread_id: str,
        add_chat_participants_request: Union[_models.AddChatParticipantsRequest, IO[bytes]],
        **kwargs: Any
    ) -> _models.AddChatParticipantsResult:
        """Adds thread participants to a thread. If participants already exist, no change occurs.

        Adds thread participants to a thread. If participants already exist, no change occurs.

        :param chat_thread_id: Id of the thread to add participants to. Required.
        :type chat_thread_id: str
        :param add_chat_participants_request: Thread participants to be added to the thread. Is either
         a AddChatParticipantsRequest type or a IO[bytes] type. Required.
        :type add_chat_participants_request:
         ~azure.communication.chat.models.AddChatParticipantsRequest or IO[bytes]
        :return: AddChatParticipantsResult or the result of cls(response)
        :rtype: ~azure.communication.chat.models.AddChatParticipantsResult
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        error_map = {
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
            401: lambda response: ClientAuthenticationError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            403: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            429: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            503: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = case_insensitive_dict(kwargs.pop("headers", {}) or {})
        _params = case_insensitive_dict(kwargs.pop("params", {}) or {})

        api_version: str = kwargs.pop("api_version", _params.pop("api-version", self._config.api_version))
        content_type: Optional[str] = kwargs.pop("content_type", _headers.pop("Content-Type", None))
        cls: ClsType[_models.AddChatParticipantsResult] = kwargs.pop("cls", None)

        content_type = content_type or "application/json"
        _json = None
        _content = None
        if isinstance(add_chat_participants_request, (IOBase, bytes)):
            _content = add_chat_participants_request
        else:
            _json = self._serialize.body(add_chat_participants_request, "AddChatParticipantsRequest")

        _request = build_add_chat_participants_request(
            chat_thread_id=chat_thread_id,
            api_version=api_version,
            content_type=content_type,
            json=_json,
            content=_content,
            headers=_headers,
            params=_params,
        )
        _request = _convert_request(_request)
        path_format_arguments = {
            "endpoint": self._serialize.url("self._config.endpoint", self._config.endpoint, "str", skip_quote=True),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [201]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)  # type: ignore
            raise HttpResponseError(response=response)

        deserialized = self._deserialize("AddChatParticipantsResult", pipeline_response)

        if cls:
            return cls(pipeline_response, deserialized, {})  # type: ignore

        return deserialized  # type: ignore

    @overload
    async def update_chat_thread_properties(  # pylint: disable=inconsistent-return-statements
        self,
        chat_thread_id: str,
        update_chat_thread_request: _models.UpdateChatThreadRequest,
        *,
        content_type: str = "application/merge-patch+json",
        **kwargs: Any
    ) -> None:
        """Updates a thread's properties.

        Updates a thread's properties.

        :param chat_thread_id: The id of the thread to update. Required.
        :type chat_thread_id: str
        :param update_chat_thread_request: Request payload for updating a chat thread. Required.
        :type update_chat_thread_request: ~azure.communication.chat.models.UpdateChatThreadRequest
        :keyword content_type: Body Parameter content-type. Content type parameter for JSON body.
         Default value is "application/merge-patch+json".
        :paramtype content_type: str
        :return: None or the result of cls(response)
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """

    @overload
    async def update_chat_thread_properties(  # pylint: disable=inconsistent-return-statements
        self,
        chat_thread_id: str,
        update_chat_thread_request: IO[bytes],
        *,
        content_type: str = "application/merge-patch+json",
        **kwargs: Any
    ) -> None:
        """Updates a thread's properties.

        Updates a thread's properties.

        :param chat_thread_id: The id of the thread to update. Required.
        :type chat_thread_id: str
        :param update_chat_thread_request: Request payload for updating a chat thread. Required.
        :type update_chat_thread_request: IO[bytes]
        :keyword content_type: Body Parameter content-type. Content type parameter for binary body.
         Default value is "application/merge-patch+json".
        :paramtype content_type: str
        :return: None or the result of cls(response)
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """

    @distributed_trace_async
    async def update_chat_thread_properties(  # pylint: disable=inconsistent-return-statements
        self,
        chat_thread_id: str,
        update_chat_thread_request: Union[_models.UpdateChatThreadRequest, IO[bytes]],
        **kwargs: Any
    ) -> None:
        """Updates a thread's properties.

        Updates a thread's properties.

        :param chat_thread_id: The id of the thread to update. Required.
        :type chat_thread_id: str
        :param update_chat_thread_request: Request payload for updating a chat thread. Is either a
         UpdateChatThreadRequest type or a IO[bytes] type. Required.
        :type update_chat_thread_request: ~azure.communication.chat.models.UpdateChatThreadRequest or
         IO[bytes]
        :return: None or the result of cls(response)
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        error_map = {
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
            401: lambda response: ClientAuthenticationError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            403: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            429: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            503: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = case_insensitive_dict(kwargs.pop("headers", {}) or {})
        _params = case_insensitive_dict(kwargs.pop("params", {}) or {})

        api_version: str = kwargs.pop("api_version", _params.pop("api-version", self._config.api_version))
        content_type: Optional[str] = kwargs.pop("content_type", _headers.pop("Content-Type", None))
        cls: ClsType[None] = kwargs.pop("cls", None)

        content_type = content_type or "application/merge-patch+json"
        _json = None
        _content = None
        if isinstance(update_chat_thread_request, (IOBase, bytes)):
            _content = update_chat_thread_request
        else:
            _json = self._serialize.body(update_chat_thread_request, "UpdateChatThreadRequest")

        _request = build_update_chat_thread_properties_request(
            chat_thread_id=chat_thread_id,
            api_version=api_version,
            content_type=content_type,
            json=_json,
            content=_content,
            headers=_headers,
            params=_params,
        )
        _request = _convert_request(_request)
        path_format_arguments = {
            "endpoint": self._serialize.url("self._config.endpoint", self._config.endpoint, "str", skip_quote=True),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [204]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)  # type: ignore
            raise HttpResponseError(response=response)

        if cls:
            return cls(pipeline_response, None, {})  # type: ignore

    @distributed_trace_async
    async def get_chat_thread_properties(self, chat_thread_id: str, **kwargs: Any) -> _models.ChatThreadProperties:
        """Gets a chat thread's properties.

        Gets a chat thread's properties.

        :param chat_thread_id: Id of the thread. Required.
        :type chat_thread_id: str
        :return: ChatThreadProperties or the result of cls(response)
        :rtype: ~azure.communication.chat.models.ChatThreadProperties
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        error_map = {
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
            401: lambda response: ClientAuthenticationError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            403: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            429: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            503: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = kwargs.pop("headers", {}) or {}
        _params = case_insensitive_dict(kwargs.pop("params", {}) or {})

        api_version: str = kwargs.pop("api_version", _params.pop("api-version", self._config.api_version))
        cls: ClsType[_models.ChatThreadProperties] = kwargs.pop("cls", None)

        _request = build_get_chat_thread_properties_request(
            chat_thread_id=chat_thread_id,
            api_version=api_version,
            headers=_headers,
            params=_params,
        )
        _request = _convert_request(_request)
        path_format_arguments = {
            "endpoint": self._serialize.url("self._config.endpoint", self._config.endpoint, "str", skip_quote=True),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)  # type: ignore
            raise HttpResponseError(response=response)

        deserialized = self._deserialize("ChatThreadProperties", pipeline_response)

        if cls:
            return cls(pipeline_response, deserialized, {})  # type: ignore

        return deserialized  # type: ignore

    @overload
    async def send_typing_notification(  # pylint: disable=inconsistent-return-statements
        self,
        chat_thread_id: str,
        send_typing_notification_request: Optional[_models.SendTypingNotificationRequest] = None,
        *,
        content_type: str = "application/json",
        **kwargs: Any
    ) -> None:
        """Posts a typing event to a thread, on behalf of a user.

        Posts a typing event to a thread, on behalf of a user.

        :param chat_thread_id: Id of the thread. Required.
        :type chat_thread_id: str
        :param send_typing_notification_request: Details of the typing notification request. Default
         value is None.
        :type send_typing_notification_request:
         ~azure.communication.chat.models.SendTypingNotificationRequest
        :keyword content_type: Body Parameter content-type. Content type parameter for JSON body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: None or the result of cls(response)
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """

    @overload
    async def send_typing_notification(  # pylint: disable=inconsistent-return-statements
        self,
        chat_thread_id: str,
        send_typing_notification_request: Optional[IO[bytes]] = None,
        *,
        content_type: str = "application/json",
        **kwargs: Any
    ) -> None:
        """Posts a typing event to a thread, on behalf of a user.

        Posts a typing event to a thread, on behalf of a user.

        :param chat_thread_id: Id of the thread. Required.
        :type chat_thread_id: str
        :param send_typing_notification_request: Details of the typing notification request. Default
         value is None.
        :type send_typing_notification_request: IO[bytes]
        :keyword content_type: Body Parameter content-type. Content type parameter for binary body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: None or the result of cls(response)
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """

    @distributed_trace_async
    async def send_typing_notification(  # pylint: disable=inconsistent-return-statements
        self,
        chat_thread_id: str,
        send_typing_notification_request: Optional[Union[_models.SendTypingNotificationRequest, IO[bytes]]] = None,
        **kwargs: Any
    ) -> None:
        """Posts a typing event to a thread, on behalf of a user.

        Posts a typing event to a thread, on behalf of a user.

        :param chat_thread_id: Id of the thread. Required.
        :type chat_thread_id: str
        :param send_typing_notification_request: Details of the typing notification request. Is either
         a SendTypingNotificationRequest type or a IO[bytes] type. Default value is None.
        :type send_typing_notification_request:
         ~azure.communication.chat.models.SendTypingNotificationRequest or IO[bytes]
        :return: None or the result of cls(response)
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        error_map = {
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
            401: lambda response: ClientAuthenticationError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            403: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            429: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
            503: lambda response: HttpResponseError(
                response=response, model=self._deserialize(_models.CommunicationErrorResponse, response)
            ),
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = case_insensitive_dict(kwargs.pop("headers", {}) or {})
        _params = case_insensitive_dict(kwargs.pop("params", {}) or {})

        api_version: str = kwargs.pop("api_version", _params.pop("api-version", self._config.api_version))
        content_type: Optional[str] = kwargs.pop("content_type", _headers.pop("Content-Type", None))
        cls: ClsType[None] = kwargs.pop("cls", None)

        content_type = content_type or "application/json"
        _json = None
        _content = None
        if isinstance(send_typing_notification_request, (IOBase, bytes)):
            _content = send_typing_notification_request
        else:
            if send_typing_notification_request is not None:
                _json = self._serialize.body(send_typing_notification_request, "SendTypingNotificationRequest")
            else:
                _json = None

        _request = build_send_typing_notification_request(
            chat_thread_id=chat_thread_id,
            api_version=api_version,
            content_type=content_type,
            json=_json,
            content=_content,
            headers=_headers,
            params=_params,
        )
        _request = _convert_request(_request)
        path_format_arguments = {
            "endpoint": self._serialize.url("self._config.endpoint", self._config.endpoint, "str", skip_quote=True),
        }
        _request.url = self._client.format_url(_request.url, **path_format_arguments)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)  # type: ignore
            raise HttpResponseError(response=response)

        if cls:
            return cls(pipeline_response, None, {})  # type: ignore



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

from azure.identity import DefaultAzureCredential
from azure.mgmt.agrifood import AgriFoodMgmtClient

"""
# PREREQUISITES
    pip install azure-identity
    pip install azure-mgmt-agrifood
# USAGE
    python locations_check_name_availability_already_exists.py

    Before run the sample, please set the values of the client ID, tenant ID and client secret
    of the AAD application as environment variables: AZURE_CLIENT_ID, AZURE_TENANT_ID,
    AZURE_CLIENT_SECRET. For more info about how to get the value, please see:
    https://docs.microsoft.com/azure/active-directory/develop/howto-create-service-principal-portal
"""


def main():
    client = AgriFoodMgmtClient(
        credential=DefaultAzureCredential(),
        solution_id="SOLUTION_ID",
        subscription_id="11111111-2222-3333-4444-555555555555",
    )

    response = client.locations.check_name_availability(
        body={"name": "existingaccountname", "type": "Microsoft.AgFoodPlatform/farmBeats"},
    )
    print(response)


# x-ms-original-file: specification/agrifood/resource-manager/Microsoft.AgFoodPlatform/preview/2021-09-01-preview/examples/Locations_CheckNameAvailability_AlreadyExists.json
if __name__ == "__main__":
    main()



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) Python Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------
import pytest
from devtools_testutils import recorded_by_proxy
from testpreparer import ClientDefaultValueClientTestBase, DefaultValuePreparer


@pytest.mark.skip("you may need to update the auto-generated test case before run it")
class TestDefaultValue(ClientDefaultValueClientTestBase):
    @DefaultValuePreparer()
    @recorded_by_proxy
    def test_put_model_property(self, defaultvalue_endpoint):
        client = self.create_client(endpoint=defaultvalue_endpoint)
        response = client.put_model_property(
            body={"name": "str", "retry": bool, "tier": "str", "timeout": 0},
        )

        # please add some check logic here by yourself
        # ...

    @DefaultValuePreparer()
    @recorded_by_proxy
    def test_get_operation_parameter(self, defaultvalue_endpoint):
        client = self.create_client(endpoint=defaultvalue_endpoint)
        response = client.get_operation_parameter(
            name="str",
        )

        # please add some check logic here by yourself
        # ...

    @DefaultValuePreparer()
    @recorded_by_proxy
    def test_get_path_parameter(self, defaultvalue_endpoint):
        client = self.create_client(endpoint=defaultvalue_endpoint)
        response = client.get_path_parameter(
            segment2="str",
            segment1="str",
        )

        # please add some check logic here by yourself
        # ...

    @DefaultValuePreparer()
    @recorded_by_proxy
    def test_get_header_parameter(self, defaultvalue_endpoint):
        client = self.create_client(endpoint=defaultvalue_endpoint)
        response = client.get_header_parameter()

        # please add some check logic here by yourself
        # ...



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) Python Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------
import pytest
from azure.mgmt.authorization import AuthorizationManagementClient

from devtools_testutils import AzureMgmtRecordedTestCase, RandomNameResourceGroupPreparer, recorded_by_proxy

AZURE_LOCATION = "eastus"


@pytest.mark.skip("you may need to update the auto-generated test case before run it")
class TestAuthorizationManagementAccessReviewHistoryDefinitionOperations(AzureMgmtRecordedTestCase):
    def setup_method(self, method):
        self.client = self.create_mgmt_client(AuthorizationManagementClient)

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_access_review_history_definition_create(self, resource_group):
        response = self.client.access_review_history_definition.create(
            history_definition_id="str",
            properties={
                "createdBy": {
                    "principalId": "str",
                    "principalName": "str",
                    "principalType": "str",
                    "userPrincipalName": "str",
                },
                "createdDateTime": "2020-02-20 00:00:00",
                "decisions": ["str"],
                "displayName": "str",
                "instances": [
                    {
                        "id": "str",
                        "name": "str",
                        "properties": {
                            "displayName": "str",
                            "downloadUri": "str",
                            "expiration": "2020-02-20 00:00:00",
                            "fulfilledDateTime": "2020-02-20 00:00:00",
                            "reviewHistoryPeriodEndDateTime": "2020-02-20 00:00:00",
                            "reviewHistoryPeriodStartDateTime": "2020-02-20 00:00:00",
                            "runDateTime": "2020-02-20 00:00:00",
                            "status": "str",
                        },
                        "type": "str",
                    }
                ],
                "reviewHistoryPeriodEndDateTime": "2020-02-20 00:00:00",
                "reviewHistoryPeriodStartDateTime": "2020-02-20 00:00:00",
                "scopes": [
                    {
                        "assignmentState": "str",
                        "excludeResourceId": "str",
                        "excludeRoleDefinitionId": "str",
                        "expandNestedMemberships": bool,
                        "inactiveDuration": "1 day, 0:00:00",
                        "includeAccessBelowResource": bool,
                        "includeInheritedAccess": bool,
                        "principalType": "str",
                        "resourceId": "str",
                        "roleDefinitionId": "str",
                    }
                ],
                "settings": {
                    "pattern": {"interval": 0, "type": "str"},
                    "range": {
                        "endDate": "2020-02-20 00:00:00",
                        "numberOfOccurrences": 0,
                        "startDate": "2020-02-20 00:00:00",
                        "type": "str",
                    },
                },
                "status": "str",
            },
            api_version="2021-12-01-preview",
        )

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_access_review_history_definition_delete_by_id(self, resource_group):
        response = self.client.access_review_history_definition.delete_by_id(
            history_definition_id="str",
            api_version="2021-12-01-preview",
        )

        # please add some check logic here by yourself
        # ...



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------
import pytest
from azure.mgmt.applicationinsights.aio import ApplicationInsightsManagementClient

from devtools_testutils import AzureMgmtRecordedTestCase, RandomNameResourceGroupPreparer
from devtools_testutils.aio import recorded_by_proxy_async

AZURE_LOCATION = "eastus"


@pytest.mark.skip("you may need to update the auto-generated test case before run it")
class TestApplicationInsightsManagementWebTestsOperationsAsync(AzureMgmtRecordedTestCase):
    def setup_method(self, method):
        self.client = self.create_mgmt_client(ApplicationInsightsManagementClient, is_async=True)

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy_async
    async def test_web_tests_list_by_resource_group(self, resource_group):
        response = self.client.web_tests.list_by_resource_group(
            resource_group_name=resource_group.name,
            api_version="2022-06-15",
        )
        result = [r async for r in response]
        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy_async
    async def test_web_tests_get(self, resource_group):
        response = await self.client.web_tests.get(
            resource_group_name=resource_group.name,
            web_test_name="str",
            api_version="2022-06-15",
        )

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy_async
    async def test_web_tests_create_or_update(self, resource_group):
        response = await self.client.web_tests.create_or_update(
            resource_group_name=resource_group.name,
            web_test_name="str",
            web_test_definition={
                "location": "str",
                "Configuration": {"WebTest": "str"},
                "Description": "str",
                "Enabled": bool,
                "Frequency": 300,
                "Kind": "str",
                "Locations": [{"Id": "str"}],
                "Name": "str",
                "Request": {
                    "FollowRedirects": bool,
                    "Headers": [{"key": "str", "value": "str"}],
                    "HttpVerb": "str",
                    "ParseDependentRequests": bool,
                    "RequestBody": "str",
                    "RequestUrl": "str",
                },
                "RetryEnabled": bool,
                "SyntheticMonitorId": "str",
                "Timeout": 30,
                "ValidationRules": {
                    "ContentValidation": {"ContentMatch": "str", "IgnoreCase": bool, "PassIfTextFound": bool},
                    "ExpectedHttpStatusCode": 0,
                    "IgnoreHttpStatusCode": bool,
                    "SSLCertRemainingLifetimeCheck": 0,
                    "SSLCheck": bool,
                },
                "id": "str",
                "kind": "str",
                "name": "str",
                "provisioningState": "str",
                "tags": {"str": "str"},
                "type": "str",
            },
            api_version="2022-06-15",
        )

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy_async
    async def test_web_tests_update_tags(self, resource_group):
        response = await self.client.web_tests.update_tags(
            resource_group_name=resource_group.name,
            web_test_name="str",
            web_test_tags={"tags": {"str": "str"}},
            api_version="2022-06-15",
        )

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy_async
    async def test_web_tests_delete(self, resource_group):
        response = await self.client.web_tests.delete(
            resource_group_name=resource_group.name,
            web_test_name="str",
            api_version="2022-06-15",
        )

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy_async
    async def test_web_tests_list(self, resource_group):
        response = self.client.web_tests.list(
            api_version="2022-06-15",
        )
        result = [r async for r in response]
        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy_async
    async def test_web_tests_list_by_component(self, resource_group):
        response = self.client.web_tests.list_by_component(
            component_name="str",
            resource_group_name=resource_group.name,
            api_version="2022-06-15",
        )
        result = [r async for r in response]
        # please add some check logic here by yourself
        # ...



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------

"""Customize generated code here.
Follow our quickstart for examples: https://aka.ms/azsdk/python/dpcodegen/python/customize
"""
from typing import List, Union

from azure.core.credentials import TokenCredential, AzureKeyCredential

from ._shared.auth_policy_utils import get_authentication_policy
from ._shared.utils import parse_connection_str
from ._client import EmailClient as EmailClientGenerated
from ._version import VERSION
from ._api_versions import DEFAULT_VERSION


class EmailClient(EmailClientGenerated):
    """A client to interact with the AzureCommunicationService Email gateway.

    This client provides operations to send an email and monitor its status.

    :param str endpoint: The endpoint url for Azure Communication Service resource.
    :param credential: The credential we use to authenticate against the service.
    :paramtype credential: ~azure.core.credentials.TokenCredential or ~azure.core.credentials.AzureKeyCredential
    :keyword api_version: Azure Communication Email API version.
        Default value is "2025-09-01".
        Note that overriding this default value may result in unsupported behavior.
    :paramtype api_version: str
    """

    def __init__(self, endpoint: str, credential: Union[TokenCredential, AzureKeyCredential], **kwargs) -> None:
        try:
            if not endpoint.lower().startswith("http"):
                endpoint = "https://" + endpoint
        except AttributeError:
            raise ValueError("Account URL must be a string.") from None
        if endpoint.endswith("/"):
            endpoint = endpoint[:-1]
        self._api_version = kwargs.pop("api_version", DEFAULT_VERSION)
        authentication_policy = get_authentication_policy(endpoint, credential)

        super().__init__(
            endpoint=endpoint,
            authentication_policy=authentication_policy,
            sdk_moniker=f"communication-email/{VERSION}",
            api_version=self._api_version,
            **kwargs,
        )

    @classmethod
    def from_connection_string(cls, conn_str: str, **kwargs) -> "EmailClient":
        """Create EmailClient from a Connection String.

        :param str conn_str: A connection string to an Azure Communication Service resource.
        :returns: Instance of EmailClient.
        :rtype: ~azure.communication.email.EmailClient
        """
        endpoint, access_key = parse_connection_str(conn_str)
        return cls(endpoint, AzureKeyCredential(access_key), **kwargs)


__all__: List[str] = ["EmailClient"]  # Add all objects you want publicly available to users at this package level


def patch_sdk():
    """Do not remove from this file.
    `patch_sdk` is a last resort escape hatch that allows you to do customizations
    you can't accomplish using the techniques described in
    https://aka.ms/azsdk/python/dpcodegen/python/customize
    """



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import os
import logging
import shlex
import time
import signal
import platform
import shutil
import tarfile
from typing import Optional, Generator, Any
import zipfile

import certifi
from dotenv import load_dotenv, find_dotenv
import pytest
import subprocess
from urllib.parse import urlparse
from urllib3.exceptions import SSLError

from ci_tools.variables import in_ci

from .config import PROXY_URL
from .fake_credentials import SANITIZED
from .helpers import get_http_client, is_live_and_not_recording
from .sanitizers import (
    add_batch_sanitizers,
    Sanitizer,
    set_custom_default_matcher,
    remove_batch_sanitizers,
)


load_dotenv(find_dotenv())

# Raise urllib3's exposed logging level so that we don't see tons of warnings while polling the proxy's availability
logging.getLogger("urllib3.connectionpool").setLevel(logging.ERROR)

_LOGGER = logging.getLogger()

CONTAINER_STARTUP_TIMEOUT = 60
PROXY_MANUALLY_STARTED = os.getenv("PROXY_MANUAL_START", False)

PROXY_CHECK_URL = PROXY_URL() + "/Info/Available"
TOOL_ENV_VAR = "PROXY_PID"

AVAILABLE_TEST_PROXY_BINARIES = {
    "Windows": {
        "AMD64": {
            "system": "Windows",
            "machine": "AMD64",
            "file_name": "test-proxy-standalone-win-x64.zip",
            "executable": "Azure.Sdk.Tools.TestProxy.exe",
        },
    },
    "Linux": {
        "X86_64": {
            "system": "Linux",
            "machine": "X86_64",
            "file_name": "test-proxy-standalone-linux-x64.tar.gz",
            "executable": "Azure.Sdk.Tools.TestProxy",
        },
        "ARM64": {
            "system": "Linux",
            "machine": "ARM64",
            "file_name": "test-proxy-standalone-linux-arm64.tar.gz",
            "executable": "Azure.Sdk.Tools.TestProxy",
        },
        "AARCH64": {  # AARCH64 and ARM64 are synonyms; use the ARM binary
            "system": "Linux",
            "machine": "AARCH64",
            "file_name": "test-proxy-standalone-linux-arm64.tar.gz",
            "executable": "Azure.Sdk.Tools.TestProxy",
        },
    },
    "Darwin": {
        "X86_64": {
            "system": "Darwin",
            "machine": "X86_64",
            "file_name": "test-proxy-standalone-osx-x64.zip",
            "executable": "Azure.Sdk.Tools.TestProxy",
        },
        "ARM64": {
            "system": "Darwin",
            "machine": "ARM64",
            "file_name": "test-proxy-standalone-osx-arm64.zip",
            "executable": "Azure.Sdk.Tools.TestProxy",
        },
    },
}

PROXY_DOWNLOAD_URL = "https://github.com/Azure/azure-sdk-tools/releases/download/Azure.Sdk.Tools.TestProxy_{}/{}"

discovered_roots = []


def _get_proxy_log_suffix() -> str:
    """Derive a log suffix based on the configured proxy port."""
    proxy_url = PROXY_URL()
    normalized = proxy_url if "://" in proxy_url else f"http://{proxy_url}"
    try:
        parsed = urlparse(normalized)
    except Exception as exc:  # pragma: no cover - defensive parsing guard
        _LOGGER.debug("Unable to parse PROXY_URL '%s': %s", proxy_url, exc)
        return "default"

    if parsed.port:
        return str(parsed.port)

    if parsed.netloc and ":" in parsed.netloc:
        candidate = parsed.netloc.rsplit(":", 1)[-1]
        if candidate.isdigit():
            return candidate

    return "default"


def get_target_version(repo_root: str) -> str:
    """Gets the target test-proxy version from the target_version.txt file in /eng/common/testproxy"""
    version_file_location = os.path.relpath("eng/common/testproxy/target_version.txt")
    override_version_file_location = os.path.relpath("eng/target_proxy_version.txt")

    version_file_location_from_root = os.path.abspath(os.path.join(repo_root, version_file_location))
    override_version_file_location_from_root = os.path.abspath(os.path.join(repo_root, override_version_file_location))

    if os.path.exists(override_version_file_location_from_root):
        with open(override_version_file_location_from_root, "r") as f:
            target_version = f.read().strip()
    else:
        with open(version_file_location_from_root, "r") as f:
            target_version = f.read().strip()

    return target_version


def get_downloaded_version(repo_root: str) -> Optional[str]:
    """Gets version from downloaded_version.txt within the local download folder"""

    downloaded_version_file = os.path.abspath(os.path.join(repo_root, ".proxy", "downloaded_version.txt"))

    if os.path.exists(downloaded_version_file):
        with open(downloaded_version_file, "r") as f:
            version = f.read().strip()
            return version
    else:
        return None


def ascend_to_root(start_dir_or_file: str) -> str:
    """Given a path, ascend until encountering a folder with a `.git` folder present within it. Return that directory.

    :param str start_dir_or_file: The starting directory or file. Either is acceptable.
    """
    if os.path.isfile(start_dir_or_file):
        current_dir = os.path.dirname(start_dir_or_file)
    else:
        current_dir = start_dir_or_file

    while current_dir is not None and not (os.path.dirname(current_dir) == current_dir):
        possible_root = os.path.join(current_dir, ".git")

        # we need the git check to prevent ascending out of the repo
        if os.path.exists(possible_root):
            if current_dir not in discovered_roots:
                discovered_roots.append(current_dir)
            return current_dir
        else:
            current_dir = os.path.dirname(current_dir)

    raise Exception(f'Requested target "{start_dir_or_file}" does not exist within a git repo.')


def check_availability() -> int:
    """Attempts request to /Info/Available. If a test-proxy instance is responding, we should get a response."""
    try:
        http_client = get_http_client(raise_on_status=False)
        response = http_client.request(method="GET", url=PROXY_CHECK_URL, timeout=10)
        return response.status
    # We get an SSLError if the container is started but the endpoint isn't available yet
    except SSLError as sslError:
        _LOGGER.debug(sslError)
        return 404
    except Exception as e:
        _LOGGER.debug(e)
        return 404


def check_certificate_location(repo_root: str) -> None:
    """Checks for a certificate bundle containing the test proxy's self-signed certificate.

    If a certificate bundle either isn't present or doesn't contain the correct test proxy certificate, a bundle is
    automatically created. SSL_CERT_DIR and REQUESTS_CA_BUNDLE are set to point to this bundle for the session.
    """

    existing_root_pem = certifi.where()
    local_dev_cert = os.path.abspath(os.path.join(repo_root, "eng", "common", "testproxy", "dotnet-devcert.crt"))
    combined_filename = os.path.basename(local_dev_cert).split(".")[0] + ".pem"
    combined_folder = os.path.join(repo_root, ".certificate")
    combined_location = os.path.join(combined_folder, combined_filename)

    # If no local certificate folder exists, create one
    if not os.path.exists(combined_folder):
        _LOGGER.info("Missing a test proxy certificate under azure-sdk-for-python/.certificate. Creating one now.")
        os.mkdir(combined_folder)

    def write_dev_cert_bundle():
        """Creates a certificate bundle with the test proxy certificate, followed by the user's existing CA bundle."""
        _LOGGER.info("Writing latest test proxy certificate to local certificate bundle.")
        # Copy the dev cert's content into the new certificate bundle
        with open(local_dev_cert, "r") as f:
            data = f.read()
        with open(combined_location, "w") as f:
            f.write(data)

        # Copy the existing CA bundle contents into the repository's certificate bundle
        with open(existing_root_pem, "r") as f:
            content = f.readlines()
        with open(combined_location, "a") as f:
            f.writelines(content)

    # If the certificate bundle isn't set up, set it up. If the bundle is present, make sure that it starts with the
    # correct certificate from eng/common/testproxy/dotnet-devcert.crt (to account for certificate rotation)
    if not os.path.exists(combined_location):
        write_dev_cert_bundle()
    else:
        with open(local_dev_cert, "r") as f:
            repo_cert = f.read()
        with open(combined_location, "r") as f:
            # The bundle should start with the test proxy's cert; only read as far in as the cert's length
            bundle_data = f.read(len(repo_cert))
        if repo_cert != bundle_data:
            write_dev_cert_bundle()

    _LOGGER.info(
        "Setting SSL_CERT_DIR, SSL_CERT_FILE, and REQUESTS_CA_BUNDLE environment variables for the current session.\n"
        f"SSL_CERT_DIR={combined_folder}\n"
        f"SSL_CERT_FILE=REQUESTS_CA_BUNDLE={combined_location}"
    )
    os.environ["SSL_CERT_DIR"] = combined_folder
    os.environ["SSL_CERT_FILE"] = combined_location
    os.environ["REQUESTS_CA_BUNDLE"] = combined_location


def check_proxy_availability() -> None:
    """Waits for the availability of the test-proxy."""
    start = time.time()
    now = time.time()
    status_code = 0
    while now - start < CONTAINER_STARTUP_TIMEOUT and status_code != 200:
        status_code = check_availability()
        now = time.time()


def prepare_local_tool(repo_root: str) -> str:
    """Returns the path to a downloaded executable."""

    target_proxy_version = get_target_version(repo_root)

    download_folder = os.path.join(repo_root, ".proxy")

    system = platform.system()  # Darwin, Linux, Windows
    machine = platform.machine().upper()  # arm64, x86_64, AMD64

    if system in AVAILABLE_TEST_PROXY_BINARIES:
        available_for_system = AVAILABLE_TEST_PROXY_BINARIES[system]

        if machine in available_for_system:
            target_info = available_for_system[machine]

            downloaded_version = get_downloaded_version(repo_root)
            download_necessary = not downloaded_version == target_proxy_version

            if download_necessary:
                if os.path.exists(download_folder):
                    # cleanup the directory for re-download
                    shutil.rmtree(download_folder)
                os.makedirs(download_folder)

                download_url = PROXY_DOWNLOAD_URL.format(target_proxy_version, target_info["file_name"])
                download_file = os.path.join(download_folder, target_info["file_name"])

                http_client = get_http_client()
                with open(download_file, "wb") as out:
                    r = http_client.request("GET", download_url, preload_content=False)
                    shutil.copyfileobj(r, out)

                if download_file.endswith(".zip"):
                    with zipfile.ZipFile(download_file, "r") as zip_ref:
                        zip_ref.extractall(download_folder)

                if download_file.endswith(".tar.gz"):
                    with tarfile.open(download_file) as tar_ref:
                        tar_ref.extractall(download_folder)

                os.remove(download_file)  # Remove downloaded file after contents are extracted

                # Record downloaded version for later comparison with target version in repo
                with open(os.path.join(download_folder, "downloaded_version.txt"), "w") as f:
                    f.writelines([target_proxy_version])

            executable_path = os.path.join(download_folder, target_info["executable"])
            # Mark the executable file as executable by all users; Mac drops these permissions during extraction
            if system == "Darwin":
                os.chmod(executable_path, 0o755)
            return os.path.abspath(executable_path).replace("\\", "/")
        else:
            _LOGGER.error(f'There are no available standalone proxy binaries for platform "{machine}".')
            raise Exception(
                "Unable to download a compatible standalone proxy for the current platform. File an issue against "
                "Azure/azure-sdk-tools with this error."
            )
    else:
        _LOGGER.error(f'There are no available standalone proxy binaries for system "{system}".')
        raise Exception(
            "Unable to download a compatible standalone proxy for the current system. File an issue against "
            "Azure/azure-sdk-tools with this error."
        )


def set_common_sanitizers() -> None:
    """Register sanitizers that will apply to all recordings throughout the SDK."""
    batch_sanitizers = {}

    # Remove headers from recordings if we don't need them, and ignore them if present
    # Authorization, for example, can contain sensitive info and can cause matching failures during challenge auth
    headers_to_ignore = "Authorization, x-ms-client-request-id, x-ms-request-id, Accept-Encoding"
    set_custom_default_matcher(excluded_headers=headers_to_ignore)
    batch_sanitizers[Sanitizer.REMOVE_HEADER] = [{"headers": headers_to_ignore}]

    # Remove OAuth interactions, which can contain client secrets and aren't necessary for playback testing
    # TODO: Determine why this breaks some test playbacks. Since sensitive info in OAuth responses is sanitized
    # through other sanitizers, it's fine to keep this off for now.
    # batch_sanitizers[Sanitizer.OAUTH_RESPONSE] = [None]

    # Body key sanitizers for sensitive fields in JSON requests/responses
    batch_sanitizers[Sanitizer.BODY_KEY] = []

    # Body regex sanitizers for sensitive patterns in request/response bodies
    batch_sanitizers[Sanitizer.BODY_REGEX] = []

    # General regex sanitizers for sensitive patterns throughout interactions
    batch_sanitizers[Sanitizer.GENERAL_REGEX] = [
        {"regex": '(?:[\\?&](sig|se|st|sv)=)(?<secret>[^&\\"\\s]*)', "group_for_replace": "secret", "value": SANITIZED},
    ]

    # Header regex sanitizers for sensitive patterns in request/response headers
    batch_sanitizers[Sanitizer.HEADER_REGEX] = []

    # URI regex sanitizers for sensitive patterns in request/response URLs
    batch_sanitizers[Sanitizer.URI_REGEX] = []

    # Send all the above sanitizers to the test proxy in a single, batch request
    add_batch_sanitizers(sanitizers=batch_sanitizers)

    # Remove certain sanitizers that are too aggressive and cause excessive playback failures
    #  - AZSDK2030: operation-location
    remove_batch_sanitizers(["AZSDK2030"])


def start_test_proxy(request) -> None:
    """Starts the test proxy and returns when the proxy server is ready to receive requests.

    In regular use cases, this will auto-start the test-proxy docker container. In CI, or when environment variable
    TF_BUILD is set, this function will start the test-proxy .NET tool.
    """

    repo_root = ascend_to_root(request.node.items[0].module.__file__)
    requires_https = PROXY_URL().startswith("https://")

    if requires_https:
        check_certificate_location(repo_root)

    if not PROXY_MANUALLY_STARTED:
        if check_availability() == 200:
            _LOGGER.debug("Tool is responding, exiting...")
        else:
            root = os.getenv("BUILD_SOURCESDIRECTORY", repo_root)
            _LOGGER.info("{} is calculated repo root".format(root))

            # If we're in CI, allow for tox environment parallelization and write proxy output to a log file
            log = None
            if in_ci():
                log_suffix = _get_proxy_log_suffix()
                log = open(os.path.join(root, f"_proxy_log_{log_suffix}.log"), "a")

            tool_name = prepare_local_tool(root)

            if requires_https:
                # Always start the proxy with these two defaults set to allow SSL connection
                passenv = {
                    "ASPNETCORE_Kestrel__Certificates__Default__Path": os.path.join(
                        root, "eng", "common", "testproxy", "dotnet-devcert.pfx"
                    ),
                    "ASPNETCORE_Kestrel__Certificates__Default__Password": "password",
                }
            else:
                passenv = {}

            # When the proxy is started in context of a directory, deletions of files under that directory crashes the test-proxy
            # due to how asp.net kestrel loads configuration files. We can disable this behavior by setting this environment
            # variable in env for the proxy process, which will allow us to clean up the --isolate directories without crashing
            # running proxies.
            passenv["DOTNET_HOSTBUILDER__RELOADCONFIGONCHANGE"] = "false"

            # If they are already set, override what we give the proxy with what is in os.environ
            passenv.update(os.environ)

            proc = subprocess.Popen(
                shlex.split(f'{tool_name} start --storage-location="{root}" -- --urls "{PROXY_URL()}"'),
                stdout=log or subprocess.DEVNULL,
                stderr=log or subprocess.STDOUT,
                env=passenv,
            )
            os.environ[TOOL_ENV_VAR] = str(proc.pid)

    # Wait for the proxy server to become available
    check_proxy_availability()
    set_common_sanitizers()


def stop_test_proxy() -> None:
    """Stops any running instance of the test proxy"""

    if not PROXY_MANUALLY_STARTED:
        _LOGGER.info("Stopping the test proxy tool...")

        try:
            os.kill(int(os.getenv(TOOL_ENV_VAR)), signal.SIGTERM)
        except:
            _LOGGER.debug("Unable to kill running test-proxy process.")


@pytest.fixture(scope="session")
def test_proxy(request) -> Generator[None, None, None]:
    """Pytest fixture to be used before running any tests that are recorded with the test proxy"""
    if is_live_and_not_recording():
        yield
    else:
        start_test_proxy(request)
        # Everything before this yield will be run before fixtures that invoke this one are run
        # Everything after it will be run after invoking fixtures are done executing
        yield
        stop_test_proxy()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
"""Contract tests for POST /responses/{response_id}/cancel behavior."""

from __future__ import annotations

import asyncio
import threading
from typing import Any

import pytest
from starlette.testclient import TestClient

from azure.ai.agentserver.responses import ResponsesAgentServerHost
from azure.ai.agentserver.responses._id_generator import IdGenerator
from tests._helpers import EventGate, poll_until


def _noop_response_handler(request: Any, context: Any, cancellation_signal: Any):
    """Minimal handler used to wire the hosting surface in contract tests."""

    async def _events():
        if False:  # pragma: no cover - required to keep async-generator shape.
            yield None

    return _events()


def _delayed_response_handler(request: Any, context: Any, cancellation_signal: Any):
    """Handler that keeps background execution cancellable for a short period."""

    async def _events():
        if cancellation_signal.is_set():
            return
        await asyncio.sleep(0.25)
        if cancellation_signal.is_set():
            return
        if False:  # pragma: no cover - keep async generator shape.
            yield None

    return _events()


def _cancellable_bg_response_handler(request: Any, context: Any, cancellation_signal: Any):
    """Handler that emits response.created then blocks until cancelled.

    Phase 3: response_created_signal is set on the first event, so run_background
    returns quickly with in_progress status while the task waits for cancellation.
    """

    async def _events():
        yield {
            "type": "response.created",
            "response": {
                "status": "in_progress",
                "output": [],
            },
        }
        # Block until cancellation signal is set
        while not cancellation_signal.is_set():
            await asyncio.sleep(0.01)

    return _events()


def _raising_response_handler(request: Any, context: Any, cancellation_signal: Any):
    """Handler that raises to transition a background response into failed."""

    async def _events():
        raise RuntimeError("simulated handler failure")
        if False:  # pragma: no cover - keep async generator shape.
            yield None

    return _events()


def _unknown_cancellation_response_handler(request: Any, context: Any, cancellation_signal: Any):
    """Handler that raises an unknown cancellation exception source."""

    async def _events():
        raise asyncio.CancelledError("unknown cancellation source")
        if False:  # pragma: no cover - keep async generator shape.
            yield None

    return _events()


def _incomplete_response_handler(request: Any, context: Any, cancellation_signal: Any):
    """Handler that emits an explicit incomplete terminal response event."""

    async def _events():
        yield {
            "type": "response.created",
            "response": {
                "status": "queued",
                "output": [],
            },
        }
        yield {
            "type": "response.in_progress",
            "response": {
                "status": "in_progress",
                "output": [],
            },
        }
        yield {
            "type": "response.incomplete",
            "response": {
                "status": "incomplete",
                "output": [],
            },
        }

    return _events()


def _make_blocking_sync_response_handler(started_gate: EventGate, release_gate: threading.Event):
    """Factory for a handler that holds a sync request in-flight for deterministic concurrent cancel checks."""

    def handler(request: Any, context: Any, cancellation_signal: Any):
        async def _events():
            started_gate.signal(True)
            while not release_gate.is_set():
                if cancellation_signal.is_set():
                    return
                await asyncio.sleep(0.01)
            if False:  # pragma: no cover - keep async generator shape.
                yield None

        return _events()

    return handler


def _build_client(handler: Any | None = None) -> TestClient:
    app = ResponsesAgentServerHost()
    app.response_handler(handler or _noop_response_handler)
    return TestClient(app)


def _create_background_response(client: TestClient, *, response_id: str | None = None) -> str:
    payload: dict[str, Any] = {
        "model": "gpt-4o-mini",
        "input": "hello",
        "stream": False,
        "store": True,
        "background": True,
    }
    if response_id is not None:
        payload["response_id"] = response_id

    create_response = client.post("/responses", json=payload)
    assert create_response.status_code == 200
    created_id = create_response.json().get("id")
    assert isinstance(created_id, str)
    return created_id


def _wait_for_status(client: TestClient, response_id: str, expected_status: str, *, timeout_s: float = 5.0) -> None:
    latest_status: str | None = None

    def _has_expected_status() -> bool:
        nonlocal latest_status
        get_response = client.get(f"/responses/{response_id}")
        if get_response.status_code != 200:
            return False
        latest_status = get_response.json().get("status")
        return latest_status == expected_status

    ok, failure = poll_until(
        _has_expected_status,
        timeout_s=timeout_s,
        interval_s=0.05,
        context_provider=lambda: {"response_id": response_id, "last_status": latest_status},
        label=f"wait for status={expected_status}",
    )
    assert ok, failure


def _assert_error(
    response: Any,
    *,
    expected_status: int,
    expected_type: str,
    expected_message: str | None = None,
    expected_code: str | None = None,
) -> None:
    assert response.status_code == expected_status
    payload = response.json()
    assert isinstance(payload.get("error"), dict)
    assert payload["error"].get("type") == expected_type
    if expected_message is not None:
        assert payload["error"].get("message") == expected_message
    if expected_code is not None:
        assert payload["error"].get("code") == expected_code, (
            f"Expected error.code={expected_code!r}, got {payload['error'].get('code')!r}"
        )


def test_cancel__cancels_background_response_and_clears_output() -> None:
    client = _build_client(_cancellable_bg_response_handler)

    response_id = _create_background_response(client)

    cancel_response = client.post(f"/responses/{response_id}/cancel")
    assert cancel_response.status_code == 200
    payload = cancel_response.json()
    assert payload.get("status") == "cancelled"
    assert payload.get("output") == []

    get_response = client.get(f"/responses/{response_id}")
    assert get_response.status_code == 200
    snapshot = get_response.json()
    assert snapshot.get("status") == "cancelled"
    assert snapshot.get("output") == []


def test_cancel__is_idempotent_for_already_cancelled_response() -> None:
    client = _build_client(_cancellable_bg_response_handler)

    response_id = _create_background_response(client)

    first_cancel = client.post(f"/responses/{response_id}/cancel")
    assert first_cancel.status_code == 200
    assert first_cancel.json().get("status") == "cancelled"
    assert first_cancel.json().get("output") == []

    second_cancel = client.post(f"/responses/{response_id}/cancel")
    assert second_cancel.status_code == 200
    assert second_cancel.json().get("status") == "cancelled"
    assert second_cancel.json().get("output") == []


def test_cancel__returns_400_for_completed_background_response() -> None:
    client = _build_client()
    response_id = _create_background_response(client)
    _wait_for_status(client, response_id, "completed")

    cancel_response = client.post(f"/responses/{response_id}/cancel")
    _assert_error(
        cancel_response,
        expected_status=400,
        expected_type="invalid_request_error",
        expected_message="Cannot cancel a completed response.",
        expected_code="invalid_request_error",
    )


def test_cancel__returns_failed_for_immediate_handler_failure() -> None:
    """Background POST waits for response.created; when the handler fails
    before emitting it, the POST returns 200 with status=failed.
    """

    def _raising_before_events(req: Any, ctx: Any, sig: Any):
        async def _ev():
            raise RuntimeError("simulated handler failure")
            if False:  # pragma: no cover
                yield None

        return _ev()

    client = _build_client(_raising_before_events)
    create_response = client.post(
        "/responses",
        json={"model": "gpt-4o-mini", "input": "hello", "stream": False, "store": True, "background": True},
    )
    # Background POST returns 200 — failure reflected in status, not HTTP code
    assert create_response.status_code == 200
    payload = create_response.json()
    assert payload.get("status") == "failed"


def test_cancel__unknown_cancellation_exception_is_treated_as_failed() -> None:
    """S-024: An unknown CancelledError (not from cancel signal) should be
    treated as a handler error, transitioning the response to failed."""
    client = _build_client(_unknown_cancellation_response_handler)
    response_id = _create_background_response(client)
    _wait_for_status(client, response_id, "failed")

    get_response = client.get(f"/responses/{response_id}")
    assert get_response.status_code == 200
    assert get_response.json().get("status") == "failed"


@pytest.mark.asyncio
async def test_cancel__stream_disconnect_sets_handler_cancellation_signal() -> None:
    """Non-bg streaming: client disconnect → handler generator cancelled.

    Uses a real Hypercorn server so that TCP disconnect propagates as
    asyncio.CancelledError to the streaming generator.
    """
    from tests._helpers import hypercorn_server

    handler_started = asyncio.Event()
    handler_cancelled = asyncio.Event()
    handler_completed = asyncio.Event()

    app = ResponsesAgentServerHost()

    @app.response_handler
    def _handler(request: Any, context: Any, cancellation_signal: Any):
        async def _events():
            from azure.ai.agentserver.responses.streaming._event_stream import ResponseEventStream

            stream = ResponseEventStream(
                response_id=context.response_id,
                model=getattr(request, "model", None),
            )
            yield stream.emit_created()
            yield stream.emit_in_progress()
            handler_started.set()
            msg = stream.add_output_item_message()
            yield msg.emit_added()
            tc = msg.add_text_content()
            yield tc.emit_added()
            for i in range(500):
                if cancellation_signal.is_set():
                    handler_cancelled.set()
                    break
                yield tc.emit_delta(f"chunk{i} ")
                await asyncio.sleep(0.02)
            yield tc.emit_text_done("cancelled")
            yield tc.emit_done()
            yield msg.emit_done()
            yield stream.emit_incomplete(reason="cancelled")
            handler_completed.set()

        return _events()

    async with hypercorn_server(app) as client:
        async with client.stream(
            "POST",
            "/responses",
            json={"model": "test", "input": "hi", "stream": True},
        ) as resp:
            assert resp.status_code == 200
            lines_read = 0
            async for _line in resp.aiter_lines():
                lines_read += 1
                if lines_read >= 5:
                    break
            # EXIT context → real TCP disconnect

        # Give server time to detect disconnect
        await asyncio.sleep(1.5)

        assert handler_started.is_set(), "Handler should have started"
        # The generator should have been cancelled by Hypercorn's
        # CancelledError propagation. The handler either saw cancellation_signal
        # or was killed by CancelledError before reaching the check.
        assert not handler_completed.is_set(), (
            "Handler should NOT have completed all 500 chunks — disconnect should stop it"
        )


@pytest.mark.asyncio
async def test_cancel__background_stream_disconnect_does_not_cancel_handler() -> None:
    """S-026: Background streaming disconnect does NOT cancel the handler.

    Uses a real Hypercorn server. Background execution is shielded from
    HTTP scope cancellation, so the handler runs to completion even after
    the streaming client disconnects.
    """
    from tests._helpers import hypercorn_server

    handler_started = asyncio.Event()
    handler_completed = asyncio.Event()

    app = ResponsesAgentServerHost()

    @app.response_handler
    def _handler(request: Any, context: Any, cancellation_signal: Any):
        async def _events():
            from azure.ai.agentserver.responses.streaming._event_stream import ResponseEventStream

            stream = ResponseEventStream(
                response_id=context.response_id,
                model=getattr(request, "model", None),
            )
            yield stream.emit_created()
            yield stream.emit_in_progress()
            handler_started.set()
            msg = stream.add_output_item_message()
            yield msg.emit_added()
            tc = msg.add_text_content()
            yield tc.emit_added()
            # Emit a handful of chunks then complete normally
            for i in range(10):
                yield tc.emit_delta(f"chunk{i} ")
                await asyncio.sleep(0.05)
            yield tc.emit_text_done("done")
            yield tc.emit_done()
            yield msg.emit_done()
            yield stream.emit_completed()
            handler_completed.set()

        return _events()

    async with hypercorn_server(app) as client:
        async with client.stream(
            "POST",
            "/responses",
            json={"model": "test", "input": "hi", "stream": True, "background": True},
        ) as resp:
            assert resp.status_code == 200
            lines_read = 0
            async for _line in resp.aiter_lines():
                lines_read += 1
                if lines_read >= 3:
                    break
            # EXIT context → real TCP disconnect

        # Wait for the background handler to complete
        await asyncio.wait_for(handler_completed.wait(), timeout=5.0)
        assert handler_started.is_set(), "Handler should have started"
        assert handler_completed.is_set(), "Background handler should complete normally despite client disconnect"


def test_cancel__returns_400_for_incomplete_background_response() -> None:
    client = _build_client(_incomplete_response_handler)
    response_id = _create_background_response(client)
    _wait_for_status(client, response_id, "incomplete")

    cancel_response = client.post(f"/responses/{response_id}/cancel")
    _assert_error(
        cancel_response,
        expected_status=400,
        expected_type="invalid_request_error",
        expected_message="Cannot cancel a response in terminal state.",
        expected_code="invalid_request_error",
    )


def test_cancel__returns_400_for_synchronous_response() -> None:
    client = _build_client()

    create_response = client.post(
        "/responses",
        json={
            "model": "gpt-4o-mini",
            "input": "hello",
            "stream": False,
            "store": True,
            "background": False,
        },
    )
    assert create_response.status_code == 200
    response_id = create_response.json()["id"]

    cancel_response = client.post(f"/responses/{response_id}/cancel")
    _assert_error(
        cancel_response,
        expected_status=400,
        expected_type="invalid_request_error",
        # After eager eviction the in-memory record is gone.  The provider
        # fallback loads the persisted response and checks background first (B1),
        # returning the correct "synchronous" message.
        expected_message="Cannot cancel a synchronous response.",
        expected_code="invalid_request_error",
    )


def test_cancel__returns_404_for_in_flight_synchronous_response() -> None:
    started_gate = EventGate()
    release_gate = threading.Event()
    client = _build_client(_make_blocking_sync_response_handler(started_gate, release_gate))
    response_id = IdGenerator.new_response_id()

    create_result: dict[str, Any] = {}

    def _issue_sync_create() -> None:
        try:
            create_result["response"] = client.post(
                "/responses",
                json={
                    "response_id": response_id,
                    "model": "gpt-4o-mini",
                    "input": "hello",
                    "stream": False,
                    "store": True,
                    "background": False,
                },
            )
        except Exception as exc:  # pragma: no cover - surfaced by assertions below.
            create_result["error"] = exc

    create_thread = threading.Thread(target=_issue_sync_create, daemon=True)
    create_thread.start()

    started, _ = started_gate.wait(timeout_s=2.0)
    assert started, "Expected sync create request to enter handler before cancel call"

    cancel_response = client.post(f"/responses/{response_id}/cancel")
    _assert_error(
        cancel_response,
        expected_status=404,
        expected_type="invalid_request_error",
        expected_code="invalid_request_error",
    )

    release_gate.set()
    create_thread.join(timeout=2.0)
    assert not create_thread.is_alive(), "Expected in-flight sync request to finish after release"

    thread_error = create_result.get("error")
    assert thread_error is None, str(thread_error)
    create_response = create_result.get("response")
    assert create_response is not None
    assert create_response.status_code == 200


def test_cancel__returns_404_for_unknown_response_id() -> None:
    from azure.ai.agentserver.responses._id_generator import IdGenerator

    client = _build_client()
    unknown_id = IdGenerator.new_response_id()

    cancel_response = client.post(f"/responses/{unknown_id}/cancel")
    _assert_error(
        cancel_response,
        expected_status=404,
        expected_type="invalid_request_error",
        expected_code="invalid_request_error",
    )


# ══════════════════════════════════════════════════════════
# B11: Cancel from queued / early in_progress state
# ══════════════════════════════════════════════════════════


def test_cancel__from_queued_or_early_in_progress_succeeds() -> None:
    """B11 — Cancel issued immediately after creation (queued/early in_progress) returns HTTP 200,
    status=cancelled, and output=[]."""
    client = _build_client(_cancellable_bg_response_handler)

    create_response = client.post(
        "/responses",
        json={
            "model": "gpt-4o-mini",
            "input": "hello",
            "stream": False,
            "store": True,
            "background": True,
        },
    )
    assert create_response.status_code == 200
    response_id = create_response.json()["id"]

    # Cancel immediately — response is likely queued or very early in_progress.
    cancel_response = client.post(f"/responses/{response_id}/cancel")
    assert cancel_response.status_code == 200

    _wait_for_status(client, response_id, "cancelled")

    get_response = client.get(f"/responses/{response_id}")
    assert get_response.status_code == 200
    payload = get_response.json()
    assert payload["status"] == "cancelled"
    assert payload.get("output") == [], f"output must be cleared for a cancelled response, got: {payload.get('output')}"


# ══════════════════════════════════════════════════════════
# B11 winddown: cancel timeout forces termination
# ══════════════════════════════════════════════════════════


def _stubborn_handler(request: Any, context: Any, cancellation_signal: Any):
    """Handler that ignores the cancellation signal entirely."""

    async def _events():
        yield {
            "type": "response.created",
            "response": {
                "status": "in_progress",
                "output": [],
            },
        }
        # Block forever, ignoring the cancellation signal
        while True:
            await asyncio.sleep(0.01)

    return _events()


def test_cancel__winddown_forces_termination_when_handler_ignores_signal() -> None:
    """B11 winddown — if the handler ignores the cancellation signal, the
    background task still reaches a terminal state within the winddown timeout."""
    import azure.ai.agentserver.responses.hosting._orchestrator as _orch

    original = _orch._CANCEL_WINDDOWN_TIMEOUT
    _orch._CANCEL_WINDDOWN_TIMEOUT = 0.5  # shorten for test speed
    try:
        client = _build_client(_stubborn_handler)
        response_id = _create_background_response(client)

        cancel_response = client.post(f"/responses/{response_id}/cancel")
        assert cancel_response.status_code == 200

        # The background task should terminate within the winddown timeout
        _wait_for_status(client, response_id, "cancelled", timeout_s=5.0)
    finally:
        _orch._CANCEL_WINDDOWN_TIMEOUT = original


# ══════════════════════════════════════════════════════════
# B12: Cancel fallback for stored terminal responses after restart
# ══════════════════════════════════════════════════════════


def test_cancel__provider_fallback_returns_400_for_completed_after_restart() -> None:
    """B12 — cancel on a completed response whose runtime record was lost
    (simulated restart) returns 400 instead of 404."""
    from azure.ai.agentserver.responses.store._memory import InMemoryResponseProvider

    provider = InMemoryResponseProvider()

    # First app instance: create and complete a response
    app1 = ResponsesAgentServerHost(store=provider)
    app1.response_handler(_noop_response_handler)
    client1 = TestClient(app1)
    response_id = _create_background_response(client1)
    _wait_for_status(client1, response_id, "completed")

    # Second app instance (simulating restart): fresh runtime state, same provider
    app2 = ResponsesAgentServerHost(store=provider)
    app2.response_handler(_noop_response_handler)
    client2 = TestClient(app2)

    cancel_response = client2.post(f"/responses/{response_id}/cancel")
    _assert_error(
        cancel_response,
        expected_status=400,
        expected_type="invalid_request_error",
        expected_message="Cannot cancel a completed response.",
        expected_code="invalid_request_error",
    )


def test_cancel__provider_fallback_returns_400_for_failed_after_restart() -> None:
    """B12 — cancel on a failed response whose runtime record was lost returns 400."""
    from azure.ai.agentserver.responses.store._memory import InMemoryResponseProvider

    provider = InMemoryResponseProvider()

    # First app instance: create a response that fails
    app1 = ResponsesAgentServerHost(store=provider)
    app1.response_handler(_raising_response_handler)
    client1 = TestClient(app1)
    response_id = _create_background_response(client1)
    _wait_for_status(client1, response_id, "failed")

    # Second app instance (simulating restart)
    app2 = ResponsesAgentServerHost(store=provider)
    app2.response_handler(_noop_response_handler)
    client2 = TestClient(app2)

    cancel_response = client2.post(f"/responses/{response_id}/cancel")
    _assert_error(
        cancel_response,
        expected_status=400,
        expected_type="invalid_request_error",
        expected_message="Cannot cancel a failed response.",
        expected_code="invalid_request_error",
    )


def test_cancel__persisted_state_is_cancelled_even_when_handler_completes_after_timeout() -> None:
    """B11 race condition: handler eventually yields response.completed after cancel.

    The durable store must still reflect 'cancelled', not 'completed'.
    """
    from azure.ai.agentserver.responses.store._memory import InMemoryResponseProvider

    provider = InMemoryResponseProvider()

    def _uncooperative_handler(request: Any, context: Any, cancellation_signal: Any):
        """Handler that ignores cancellation and eventually completes."""

        async def _events():
            yield {
                "type": "response.created",
                "response": {"status": "in_progress", "output": []},
            }
            # Deliberately ignores cancellation_signal — simulates uncooperative handler.
            # The short sleep ensures the handler is still "running" when cancel comes in,
            # but completes before the 10s winddown deadline.
            await asyncio.sleep(0.5)
            yield {
                "type": "response.completed",
                "response": {"status": "completed", "output": []},
            }

        return _events()

    app = ResponsesAgentServerHost(store=provider)
    app.response_handler(_uncooperative_handler)
    client = TestClient(app)

    create = client.post(
        "/responses",
        json={"model": "test", "input": "hello", "stream": False, "store": True, "background": True},
    )
    assert create.status_code == 200
    response_id = create.json()["id"]

    # Cancel — the handler is still running
    cancel = client.post(f"/responses/{response_id}/cancel")
    assert cancel.status_code == 200
    assert cancel.json()["status"] == "cancelled"

    # Wait for background task to fully finalize
    import time

    time.sleep(2.0)

    # GET from durable store must show cancelled
    get = client.get(f"/responses/{response_id}")
    assert get.status_code == 200
    assert get.json()["status"] == "cancelled", (
        "B11: Persisted state must be 'cancelled' — cancellation always wins, "
        "even when the handler yields response.completed after cancel"
    )


def test_cancel__in_progress_response_triggers_cancellation_signal() -> None:
    """Cancel triggers the cancellation_signal provided to the handler.

    Ported from CancelResponseProtocolTests.Cancel_InProgressResponse_TriggersCancellationToken.
    """

    def _tracking_handler(request: Any, context: Any, cancellation_signal: Any):
        async def _events():
            yield {
                "type": "response.created",
                "response": {"status": "in_progress", "output": []},
            }
            # Block until cancel; the asyncio.sleep yields to the event loop
            # so the cancel endpoint's signal actually propagates.
            for _ in range(500):
                if cancellation_signal.is_set():
                    return
                await asyncio.sleep(0.01)

        return _events()

    client = _build_client(_tracking_handler)

    create = client.post(
        "/responses",
        json={"model": "test", "input": "hello", "stream": False, "store": True, "background": True},
    )
    assert create.status_code == 200
    response_id = create.json()["id"]

    cancel = client.post(f"/responses/{response_id}/cancel")
    assert cancel.status_code == 200

    # If the signal was triggered the handler should have exited and the
    # response reached a terminal state ("cancelled").
    _wait_for_status(client, response_id, "cancelled", timeout_s=5.0)



class AuthoringTestHelper:
    """Utility helper for creating and exporting authoring test projects."""

    @staticmethod
    def create_test_project(
        client,
        project_name="IsaacNewton",
        is_deployable=False,
        add_sources=False,
        get_export_url=False,
        delete_old_project=False,
        **kwargs,
    ):
        # Separate polling keyword (if present) from feature flags to avoid accidental binding confusions
        polling_interval = kwargs.pop("polling_interval", None)
        client.create_project(
            project_name=project_name,
            options={
                "description": "Biography of Sir Isaac Newton",
                "language": "en",
                "multilingualResource": True,
                "settings": {"defaultAnswer": "no answer"},
            },
        )
        source_kwargs = {"polling_interval": polling_interval} if polling_interval is not None else {}
        if is_deployable or add_sources:
            AuthoringTestHelper.add_sources(client, project_name, **source_kwargs)
        if get_export_url:
            return AuthoringTestHelper.export_project(
                client, project_name, delete_project=delete_old_project, **source_kwargs
            )
        return None

    @staticmethod
    def add_sources(client, project_name, **kwargs):
        poller = client.begin_update_sources(
            project_name=project_name,
            sources=[
                {
                    "op": "add",
                    "value": {
                        "displayName": "Isaac Newton Bio",
                        "sourceUri": "https://wikipedia.org/wiki/Isaac_Newton",
                        "sourceKind": "url",
                    },
                }
            ],
            **kwargs,
        )
        poller.result()

    @staticmethod
    def export_project(client, project_name, delete_project=True, **kwargs): # pylint: disable=useless-return
        # begin_export poller is typed as LROPoller[None]; generator currently discards
        # the final body so result() returns None. We only validate successful completion.
        export_poller = client.begin_export(project_name=project_name, file_format="json", **kwargs)
        export_poller.result()  # ensure completion (raises on failure)
        if delete_project:
            delete_poller = client.begin_delete_project(project_name=project_name, **kwargs)
            delete_poller.result()
        return None

class AuthoringAsyncTestHelper:
    """Async utility helper for creating and exporting authoring test projects."""

    @staticmethod
    async def create_test_project(
        client,
        project_name="IsaacNewton",
        is_deployable=False,
        add_sources=False,
        get_export_url=False,
        delete_old_project=False,
        **kwargs,
    ):
        polling_interval = kwargs.pop("polling_interval", None)
        await client.create_project(
            project_name=project_name,
            options={
                "description": "Biography of Sir Isaac Newton",
                "language": "en",
                "multilingualResource": True,
                "settings": {"defaultAnswer": "no answer"},
            },
        )
        source_kwargs = {"polling_interval": polling_interval} if polling_interval is not None else {}
        if is_deployable or add_sources:
            await AuthoringAsyncTestHelper.add_sources(client, project_name, **source_kwargs)
        if get_export_url:
            return await AuthoringAsyncTestHelper.export_project(
                client, project_name, delete_project=delete_old_project, **source_kwargs
            )

    @staticmethod
    async def add_sources(client, project_name, **kwargs):
        poller = await client.begin_update_sources(
            project_name=project_name,
            sources=[
                {
                    "op": "add",
                    "value": {
                        "displayName": "Isaac Newton Bio",
                        "sourceUri": "https://wikipedia.org/wiki/Isaac_Newton",
                        "sourceKind": "url",
                    },
                }
            ],
            **kwargs,
        )
        await poller.result()

    @staticmethod
    async def export_project(client, project_name, delete_project=True, **kwargs):
        export_poller = await client.begin_export(project_name=project_name, file_format="json", **kwargs)
        await export_poller.result()
        if delete_project:
            delete_poller = await client.begin_delete_project(project_name=project_name, **kwargs)
            await delete_poller.result()
        return None



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
"""Function call builders: function-call and function-call-output output items."""

from __future__ import annotations

from collections.abc import AsyncIterable
from copy import deepcopy
from typing import TYPE_CHECKING, AsyncIterator, Iterator, cast

from ...models import _generated as generated_models
from ._base import BaseOutputItemBuilder, _require_non_empty

if TYPE_CHECKING:
    from .._event_stream import ResponseEventStream


class OutputItemFunctionCallBuilder(BaseOutputItemBuilder):
    """Scoped builder for a function-call output item in stream mode."""

    def __init__(
        self,
        stream: "ResponseEventStream",
        output_index: int,
        item_id: str,
        name: str,
        call_id: str,
    ) -> None:
        """Initialize the function-call output item builder.

        :param stream: The parent event stream.
        :type stream: ResponseEventStream
        :param output_index: Zero-based index of this output item.
        :type output_index: int
        :param item_id: Unique identifier for this output item.
        :type item_id: str
        :param name: The function name being called.
        :type name: str
        :param call_id: Unique identifier for this function call.
        :type call_id: str
        """
        super().__init__(stream=stream, output_index=output_index, item_id=item_id)
        self._name = _require_non_empty(name, "name")
        self._call_id = _require_non_empty(call_id, "call_id")
        self._final_arguments: str | None = None

    @property
    def name(self) -> str:
        """Return the function name.

        :returns: The function name.
        :rtype: str
        """
        return self._name

    @property
    def call_id(self) -> str:
        """Return the function call identifier.

        :returns: The call ID.
        :rtype: str
        """
        return self._call_id

    def emit_added(self) -> generated_models.ResponseOutputItemAddedEvent:
        """Emit an ``output_item.added`` event for this function call.

        :returns: The emitted event.
        :rtype: ResponseOutputItemAddedEvent
        """
        return self._emit_added(
            {
                "type": "function_call",
                "id": self._item_id,
                "call_id": self._call_id,
                "name": self._name,
                "arguments": "",
                "status": "in_progress",
            }
        )

    def emit_arguments_delta(self, delta: str) -> generated_models.ResponseFunctionCallArgumentsDeltaEvent:
        """Emit a function-call arguments delta event.

        :param delta: The incremental arguments text fragment.
        :type delta: str
        :returns: The emitted event.
        :rtype: ResponseFunctionCallArgumentsDeltaEvent
        """
        return cast(
            generated_models.ResponseFunctionCallArgumentsDeltaEvent,
            self._stream._emit_event(  # pylint: disable=protected-access
                {
                    "type": generated_models.ResponseStreamEventType.RESPONSE_FUNCTION_CALL_ARGUMENTS_DELTA.value,
                    "item_id": self._item_id,
                    "output_index": self._output_index,
                    "delta": delta,
                }
            ),
        )

    def emit_arguments_done(self, arguments: str) -> generated_models.ResponseFunctionCallArgumentsDoneEvent:
        """Emit a function-call arguments done event.

        :param arguments: The final, complete arguments string.
        :type arguments: str
        :returns: The emitted event.
        :rtype: ResponseFunctionCallArgumentsDoneEvent
        """
        self._final_arguments = arguments
        return cast(
            generated_models.ResponseFunctionCallArgumentsDoneEvent,
            self._stream._emit_event(  # pylint: disable=protected-access
                {
                    "type": generated_models.ResponseStreamEventType.RESPONSE_FUNCTION_CALL_ARGUMENTS_DONE.value,
                    "item_id": self._item_id,
                    "output_index": self._output_index,
                    "name": self._name,
                    "arguments": arguments,
                }
            ),
        )

    def emit_done(self) -> generated_models.ResponseOutputItemDoneEvent:
        """Emit an ``output_item.done`` event for this function call.

        :returns: The emitted event.
        :rtype: ResponseOutputItemDoneEvent
        """
        return self._emit_done(
            {
                "type": "function_call",
                "id": self._item_id,
                "call_id": self._call_id,
                "name": self._name,
                "arguments": self._final_arguments or "",
                "status": "completed",
            }
        )

    # ---- Sub-item convenience generators (S-053) ----

    def arguments(self, args: str) -> Iterator[generated_models.ResponseStreamEvent]:
        """Yield the argument delta and done events.

        Emits ``function_call_arguments.delta`` followed by
        ``function_call_arguments.done``.

        :param args: The complete arguments string.
        :type args: str
        :returns: An iterator of events.
        :rtype: Iterator[ResponseStreamEvent]
        """
        yield self.emit_arguments_delta(args)
        yield self.emit_arguments_done(args)

    async def aarguments(self, args: str | AsyncIterable[str]) -> AsyncIterator[generated_models.ResponseStreamEvent]:
        """Async variant of :meth:`arguments` with streaming support.

        When *args* is a string, behaves identically to :meth:`arguments`.
        When *args* is an async iterable of string chunks, emits one
        ``function_call_arguments.delta`` per chunk in real time (S-055),
        then ``function_call_arguments.done`` with the accumulated text.

        :param args: Complete arguments string or async iterable of chunks.
        :type args: str | AsyncIterable[str]
        :returns: An async iterator of events.
        :rtype: AsyncIterator[ResponseStreamEvent]
        """
        if isinstance(args, str):
            for event in self.arguments(args):
                yield event
            return
        accumulated: list[str] = []
        async for chunk in args:
            accumulated.append(chunk)
            yield self.emit_arguments_delta(chunk)
        yield self.emit_arguments_done("".join(accumulated))


class OutputItemFunctionCallOutputBuilder(BaseOutputItemBuilder):
    """Scoped builder for a function-call-output item in stream mode."""

    def __init__(
        self,
        stream: "ResponseEventStream",
        output_index: int,
        item_id: str,
        call_id: str,
    ) -> None:
        """Initialize the function-call-output item builder.

        :param stream: The parent event stream.
        :type stream: ResponseEventStream
        :param output_index: Zero-based index of this output item.
        :type output_index: int
        :param item_id: Unique identifier for this output item.
        :type item_id: str
        :param call_id: The call ID of the function call this output belongs to.
        :type call_id: str
        """
        super().__init__(stream=stream, output_index=output_index, item_id=item_id)
        self._call_id = _require_non_empty(call_id, "call_id")
        self._final_output: (
            str
            | list[
                generated_models.InputTextContentParam
                | generated_models.InputImageContentParamAutoParam
                | generated_models.InputFileContentParam
            ]
            | None
        ) = None

    @property
    def call_id(self) -> str:
        """Return the function call identifier.

        :returns: The call ID.
        :rtype: str
        """
        return self._call_id

    def emit_added(
        self,
        output: str
        | list[
            generated_models.InputTextContentParam
            | generated_models.InputImageContentParamAutoParam
            | generated_models.InputFileContentParam
        ]
        | None = None,
    ) -> generated_models.ResponseOutputItemAddedEvent:
        """Emit an ``output_item.added`` event for this function-call output.

        :param output: Optional initial output value.
        :type output: str | list[InputTextContentParam | InputImageContentParamAutoParam | InputFileContentParam] | None
        :returns: The emitted event.
        :rtype: ResponseOutputItemAddedEvent
        """
        return self._emit_added(
            {
                "type": "function_call_output",
                "id": self._item_id,
                "call_id": self._call_id,
                "output": deepcopy(output) if output is not None else "",
                "status": "in_progress",
            }
        )

    def emit_done(
        self,
        output: str
        | list[
            generated_models.InputTextContentParam
            | generated_models.InputImageContentParamAutoParam
            | generated_models.InputFileContentParam
        ]
        | None = None,
    ) -> generated_models.ResponseOutputItemDoneEvent:
        """Emit an ``output_item.done`` event for this function-call output.

        :param output: Optional final output value. Uses previously set output if ``None``.
        :type output: str | list[InputTextContentParam | InputImageContentParamAutoParam | InputFileContentParam] | None
        :returns: The emitted event.
        :rtype: ResponseOutputItemDoneEvent
        """
        if output is not None:
            self._final_output = deepcopy(output)

        return self._emit_done(
            {
                "type": "function_call_output",
                "id": self._item_id,
                "call_id": self._call_id,
                "output": deepcopy(self._final_output) if self._final_output is not None else "",
                "status": "completed",
            }
        )



# coding=utf-8
# ------------------------------------
# Copyright (c) Microsoft.
# Licensed under the MIT License.
# ------------------------------------

"""
FILE: sample_delete_trained_model.py
DESCRIPTION:
    This sample demonstrates how to delete a **trained model** from a Text Authoring project.
USAGE:
    python sample_delete_trained_model.py
REQUIRED ENV VARS (for AAD / DefaultAzureCredential):
    AZURE_TEXT_ENDPOINT
    AZURE_CLIENT_ID
    AZURE_TENANT_ID
    AZURE_CLIENT_SECRET
NOTE:
    If you want to use AzureKeyCredential instead, set:
      - AZURE_TEXT_ENDPOINT
      - AZURE_TEXT_KEY
OPTIONAL ENV VARS:
    PROJECT_NAME         # defaults to "<project-name>"
    TRAINED_MODEL_LABEL  # defaults to "<trained-model-label>"
"""

# [START text_authoring_delete_trained_model]
import os
from azure.core.exceptions import HttpResponseError
from azure.identity import DefaultAzureCredential
from azure.ai.textanalytics.authoring import TextAuthoringClient


def sample_delete_trained_model():
    # settings
    endpoint = os.environ["AZURE_TEXT_ENDPOINT"]
    project_name = os.environ.get("PROJECT_NAME", "<project-name>")
    trained_model_label = os.environ.get("TRAINED_MODEL_LABEL", "<trained-model-label>")

    # create a client with AAD
    credential = DefaultAzureCredential()
    client = TextAuthoringClient(endpoint, credential=credential)

    # project-scoped client
    project_client = client.get_project_client(project_name)

    try:
        project_client.trained_model.delete_trained_model(trained_model_label)
        print("Result: Deleted successfully")
    except HttpResponseError as e:
        print(f"Delete failed: {e.message}")


# [END text_authoring_delete_trained_model]


def main():
    sample_delete_trained_model()


if __name__ == "__main__":
    main()



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# Below are common methods for the devops build steps. This is the common location that will be updated with
# package targeting during release.
import os
import argparse
from packaging.version import parse
import logging

from ci_tools.versioning.version_shared import (
    get_packages,
    set_version_py,
    set_dev_classifier,
    update_change_log,
)
from ci_tools.variables import discover_repo_root

logging.getLogger().setLevel(logging.INFO)


def increment_version(old_version):
    parsed_version = parse(old_version)
    release = parsed_version.release

    if parsed_version.is_prerelease:
        prerelease_version = parsed_version.pre[1]
        return "{0}.{1}.{2}b{3}".format(release[0], release[1], release[2], prerelease_version + 1)

    return "{0}.{1}.{2}".format(release[0], release[1], release[2] + 1)


def version_increment_main():
    parser = argparse.ArgumentParser(
        description="Increments version for a given package name based on the released version"
    )
    parser.add_argument(
        "--package-name",
        required=True,
        help="name of package (accepts both formats: azure-service-package and azure_service_package)",
    )
    parser.add_argument(
        dest="glob_string",
        nargs="?",
        help=(
            "A comma separated list of glob strings that will target the top level directories that contain packages."
            'Examples: All = "azure-*", Single = "azure-keyvault", Targeted Multiple = "azure-keyvault,azure-mgmt-resource"'
        ),
    )
    parser.add_argument(
        "--service",
        required=True,
        help="name of the service for which to set the dev build id (e.g. keyvault)",
    )
    parser.add_argument(
        "--repo",
        default=None,
        dest="repo",
        help=(
            "Where is the start directory that we are building against? If not provided, the current working directory will be used. Please ensure you are within the azure-sdk-for-python repository."
        ),
    )

    args = parser.parse_args()
    root_dir = args.repo or discover_repo_root()

    package_name = args.package_name.replace("_", "-")

    packages = get_packages(args, package_name, additional_excludes=["mgmt", "-nspkg"], root_dir=root_dir)

    package_map = {pkg.name: pkg for pkg in packages}

    if package_name not in package_map:
        raise ValueError("Package name not found: {}".format(package_name))

    target_package = package_map[package_name]

    new_version = increment_version(target_package.version)
    print("{0}: {1} -> {2}".format(package_name, target_package.version, new_version))

    set_version_py(target_package.setup_filename, new_version)
    set_dev_classifier(target_package.setup_filename, new_version)
    update_change_log(
        target_package.setup_filename, new_version, args.service, args.package_name, True, False, root_dir=root_dir
    )



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------

"""Compatibility shim for generated model symbols."""

from .sdk.models.models._models import *  # type: ignore # noqa: F401,F403



# pylint: disable=line-too-long,useless-suppression
# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

from azure.identity import DefaultAzureCredential

from azure.mgmt.appcontainers import ContainerAppsAPIClient

"""
# PREREQUISITES
    pip install azure-identity
    pip install azure-mgmt-appcontainers
# USAGE
    python container_apps_create_or_update.py

    Before run the sample, please set the values of the client ID, tenant ID and client secret
    of the AAD application as environment variables: AZURE_CLIENT_ID, AZURE_TENANT_ID,
    AZURE_CLIENT_SECRET. For more info about how to get the value, please see:
    https://docs.microsoft.com/azure/active-directory/develop/howto-create-service-principal-portal
"""


def main():
    client = ContainerAppsAPIClient(
        credential=DefaultAzureCredential(),
        subscription_id="34adfa4f-cedf-4dc0-ba29-b6d1a69ab345",
    )

    response = client.container_apps.begin_create_or_update(
        resource_group_name="rg",
        container_app_name="testcontainerapp0",
        container_app_envelope={
            "identity": {
                "type": "SystemAssigned,UserAssigned",
                "userAssignedIdentities": {
                    "/subscriptions/34adfa4f-cedf-4dc0-ba29-b6d1a69ab345/resourcegroups/rg/providers/Microsoft.ManagedIdentity/userAssignedIdentities/myidentity": {}
                },
            },
            "location": "East US",
            "properties": {
                "configuration": {
                    "dapr": {
                        "appHealth": {
                            "enabled": True,
                            "path": "/health",
                            "probeIntervalSeconds": 3,
                            "probeTimeoutMilliseconds": 1000,
                            "threshold": 3,
                        },
                        "appPort": 3000,
                        "appProtocol": "http",
                        "enableApiLogging": True,
                        "enabled": True,
                        "httpMaxRequestSize": 10,
                        "httpReadBufferSize": 30,
                        "logLevel": "debug",
                        "maxConcurrency": 10,
                    },
                    "identitySettings": [
                        {
                            "identity": "/subscriptions/34adfa4f-cedf-4dc0-ba29-b6d1a69ab345/resourcegroups/rg/providers/Microsoft.ManagedIdentity/userAssignedIdentities/myidentity",
                            "lifecycle": "All",
                        },
                        {"identity": "system", "lifecycle": "Init"},
                    ],
                    "ingress": {
                        "additionalPortMappings": [
                            {"external": True, "targetPort": 1234},
                            {"exposedPort": 3456, "external": False, "targetPort": 2345},
                        ],
                        "clientCertificateMode": "accept",
                        "corsPolicy": {
                            "allowCredentials": True,
                            "allowedHeaders": ["HEADER1", "HEADER2"],
                            "allowedMethods": ["GET", "POST"],
                            "allowedOrigins": ["https://a.test.com", "https://b.test.com"],
                            "exposeHeaders": ["HEADER3", "HEADER4"],
                            "maxAge": 1234,
                        },
                        "customDomains": [
                            {
                                "bindingType": "SniEnabled",
                                "certificateId": "/subscriptions/34adfa4f-cedf-4dc0-ba29-b6d1a69ab345/resourceGroups/rg/providers/Microsoft.App/managedEnvironments/demokube/certificates/my-certificate-for-my-name-dot-com",
                                "name": "www.my-name.com",
                            },
                            {
                                "bindingType": "SniEnabled",
                                "certificateId": "/subscriptions/34adfa4f-cedf-4dc0-ba29-b6d1a69ab345/resourceGroups/rg/providers/Microsoft.App/managedEnvironments/demokube/certificates/my-certificate-for-my-other-name-dot-com",
                                "name": "www.my-other-name.com",
                            },
                        ],
                        "external": True,
                        "ipSecurityRestrictions": [
                            {
                                "action": "Allow",
                                "description": "Allowing all IP's within the subnet below to access containerapp",
                                "ipAddressRange": "192.168.1.1/32",
                                "name": "Allow work IP A subnet",
                            },
                            {
                                "action": "Allow",
                                "description": "Allowing all IP's within the subnet below to access containerapp",
                                "ipAddressRange": "192.168.1.1/8",
                                "name": "Allow work IP B subnet",
                            },
                        ],
                        "stickySessions": {"affinity": "sticky"},
                        "targetPort": 3000,
                        "traffic": [{"label": "production", "revisionName": "testcontainerapp0-ab1234", "weight": 100}],
                    },
                    "maxInactiveRevisions": 10,
                    "runtime": {"java": {"enableMetrics": True}},
                    "service": {"type": "redis"},
                },
                "environmentId": "/subscriptions/34adfa4f-cedf-4dc0-ba29-b6d1a69ab345/resourceGroups/rg/providers/Microsoft.App/managedEnvironments/demokube",
                "template": {
                    "containers": [
                        {
                            "image": "repo/testcontainerapp0:v1",
                            "name": "testcontainerapp0",
                            "probes": [
                                {
                                    "httpGet": {
                                        "httpHeaders": [{"name": "Custom-Header", "value": "Awesome"}],
                                        "path": "/health",
                                        "port": 8080,
                                    },
                                    "initialDelaySeconds": 3,
                                    "periodSeconds": 3,
                                    "type": "Liveness",
                                }
                            ],
                            "volumeMounts": [
                                {"mountPath": "/mnt/path1", "subPath": "subPath1", "volumeName": "azurefile"},
                                {"mountPath": "/mnt/path2", "subPath": "subPath2", "volumeName": "nfsazurefile"},
                            ],
                        }
                    ],
                    "initContainers": [
                        {
                            "args": ["-c", "while true; do echo hello; sleep 10;done"],
                            "command": ["/bin/sh"],
                            "image": "repo/testcontainerapp0:v4",
                            "name": "testinitcontainerApp0",
                            "resources": {"cpu": 0.5, "memory": "1Gi"},
                        }
                    ],
                    "scale": {
                        "cooldownPeriod": 350,
                        "maxReplicas": 5,
                        "minReplicas": 1,
                        "pollingInterval": 35,
                        "rules": [
                            {
                                "custom": {"metadata": {"concurrentRequests": "50"}, "type": "http"},
                                "name": "httpscalingrule",
                            },
                            {
                                "custom": {
                                    "identity": "/subscriptions/34adfa4f-cedf-4dc0-ba29-b6d1a69ab345/resourcegroups/rg/providers/Microsoft.ManagedIdentity/userAssignedIdentities/myidentity",
                                    "metadata": {
                                        "messageCount": "5",
                                        "namespace": "mynamespace",
                                        "queueName": "myqueue",
                                    },
                                    "type": "azure-servicebus",
                                },
                                "name": "servicebus",
                            },
                            {
                                "azureQueue": {
                                    "accountName": "account1",
                                    "identity": "system",
                                    "queueLength": 1,
                                    "queueName": "queue1",
                                },
                                "name": "azure-queue",
                            },
                        ],
                    },
                    "serviceBinds": [
                        {
                            "name": "redisService",
                            "serviceId": "/subscriptions/34adfa4f-cedf-4dc0-ba29-b6d1a69ab345/resourceGroups/rg/providers/Microsoft.App/containerApps/redisService",
                        }
                    ],
                    "volumes": [
                        {"name": "azurefile", "storageName": "storage", "storageType": "AzureFile"},
                        {"name": "nfsazurefile", "storageName": "nfsStorage", "storageType": "NfsAzureFile"},
                    ],
                },
                "workloadProfileName": "My-GP-01",
            },
        },
    ).result()
    print(response)


# x-ms-original-file: specification/app/resource-manager/Microsoft.App/ContainerApps/stable/2025-07-01/examples/ContainerApps_CreateOrUpdate.json
if __name__ == "__main__":
    main()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# This script is used to create issues for client libraries failing the vnext of mypy, pyright, and pylint.
from __future__ import annotations
from typing import Optional

import sys
import os
import subprocess
import logging
import datetime
import re
import calendar
import typing
import pathlib
from typing_extensions import Literal
from github import Github, Auth, GithubException

from ci_tools.variables import discover_repo_root

logging.getLogger().setLevel(logging.INFO)

CHECK_TYPE = Literal["mypy", "pylint", "pyright", "sphinx"]


def get_version_running(check_type: CHECK_TYPE) -> str:
    commands = [sys.executable, "-m", check_type, "--version"]
    version = subprocess.run(
        commands,
        check=True,
        capture_output=True,
    )
    version = version.stdout.rstrip().decode("utf-8")
    version_running = re.findall(r"(\d+.\d+.\d+)", version)[0]
    logging.info(f"Running {check_type} version {version_running}")
    return version_running


def get_build_link(check_type: CHECK_TYPE) -> str:
    build_id = os.getenv("BUILD_BUILDID")
    job_id = os.getenv("SYSTEM_JOBID")

    next_id: str
    if check_type == "mypy":
        next_id = "c4b2a078-69a7-55a2-d776-67715c71590f"
    if check_type == "pyright":
        next_id = "d243185e-b901-5eef-29fe-f7943e030451"
    if check_type == "pylint":
        next_id = "b33d1587-3539-5735-af43-e3e62f02ca4b"
    if check_type == "sphinx":
        next_id = "82919efa-82d6-5dc4-2e9a-f82117bff292"

    return (
        f"https://dev.azure.com/azure-sdk/internal/_build/results?buildId={build_id}&view=logs&j={job_id}&t={next_id}"
    )


def get_merge_dates(year: str) -> typing.List[datetime.datetime]:
    """We'll merge the latest version of the type checker/linter quarterly
    on the Monday after release week. This function returns those 4 Mondays
    for the given year.
    """
    c = calendar.Calendar(firstweekday=calendar.FRIDAY)
    first = c.monthdatescalendar(year, 1)
    second = c.monthdatescalendar(year, 4)
    third = c.monthdatescalendar(year, 7)
    fourth = c.monthdatescalendar(year, 10)

    merge_months = [first, second, third, fourth]

    merge_dates = []
    for month in merge_months:
        code_complete = [
            day for week in month for day in week if day.weekday() == calendar.FRIDAY and day.month in [1, 4, 7, 10]
        ][0]
        monday_after_release_week = code_complete + datetime.timedelta(days=10)
        merge_dates.append(monday_after_release_week)
    return merge_dates


def get_date_for_version_bump(today: datetime.datetime) -> str:
    merge_dates = get_merge_dates(today.year)
    try:
        merge_date = min(date for date in merge_dates if date >= today)
    except ValueError:
        # today's date is after October merge date, so rollover to next year
        merge_dates = get_merge_dates(today.year + 1)
        merge_date = min(date for date in merge_dates if date >= today)
    return merge_date.strftime("%Y-%m-%d")


def get_labels(package_name: str, service: str) -> tuple[list[str], list[str]]:
    repo_root = discover_repo_root()
    codeowners_path = pathlib.Path(repo_root) / ".github" / "CODEOWNERS"
    with open(codeowners_path, "r") as codeowners_file:
        codeowners = codeowners_file.readlines()

    label = ""
    service_label = ""
    service_owners = []
    labels = []
    assignees = []
    if "mgmt" in package_name:
        labels.append("Mgmt")
    for line in codeowners:
        if line.startswith("# PRLabel:"):
            label = line.split("# PRLabel: %")[1].strip()
        if label and line.startswith("/sdk/"):
            parts = [part for part in line.split("@")[0].split("/") if part.strip()][1:]
            if len(parts) > 2:
                continue
            try:
                service_directory = parts[0]
            except IndexError:
                # it was a single file
                continue
            try:
                library = parts[1]
                if package_name == library:
                    labels.append(label)
                    # Extract codeowners from the line
                    tokens = line.split()
                    owners = [token.lstrip("@") for token in tokens if token.startswith("@")]
                    # Filter out teams (containing '/')
                    assignees = [owner for owner in owners if "/" not in owner]
                    return labels, assignees
            except IndexError:
                if service_directory == service:
                    service_label = label
                    # Extract service-level codeowners
                    tokens = line.split()
                    owners = [token.lstrip("@") for token in tokens if token.startswith("@")]
                    service_owners = [owner for owner in owners if "/" not in owner]

    if service_label:
        labels.append(service_label)
        assignees = service_owners
    return labels, assignees


def create_vnext_issue(package_dir: str, check_type: CHECK_TYPE, check_version: Optional[str] = None) -> None:
    """This is called when a client library fails a vnext check.
    An issue is created with the details or an existing issue is updated with the latest information."""

    package_path = pathlib.Path(package_dir)
    package_name = package_path.name
    service_directory = package_path.parent.name
    auth = Auth.Token(os.environ["GH_TOKEN"])
    g = Github(auth=auth)

    today = datetime.date.today()
    repo = g.get_repo("Azure/azure-sdk-for-python")

    issues = repo.get_issues(state="open", labels=[check_type], creator="azure-sdk")
    vnext_issue = [issue for issue in issues if issue.title.split("needs")[0].strip() == package_name]

    version = check_version or get_version_running(check_type)
    build_link = get_build_link(check_type)
    merge_date = get_date_for_version_bump(today)
    error_type = "linting" if check_type == "pylint" else "docstring" if check_type == "sphinx" else "typing"
    guide_link = (
        "[Pylint Guide](https://github.com/Azure/azure-sdk-for-python/blob/main/doc/dev/pylint_checking.md)"
        if check_type == "pylint"
        else (
            "[Sphinx and docstring checker](https://github.com/Azure/azure-sdk-for-python/blob/main/doc/eng_sys_checks.md#sphinx-and-docstring-checker)"
            if check_type == "sphinx"
            else "[Typing Guide](https://github.com/Azure/azure-sdk-for-python/blob/main/doc/dev/static_type_checking.md#run-mypy)"
        )
    )

    repo_root = discover_repo_root()
    title = f"{package_name} needs {error_type} updates for {check_type} version {version}"
    template = (
        f"**ACTION NEEDED:** This version of {check_type} will be merged on **{merge_date}**. "
        f"The build will begin to fail for this library if errors are not fixed."
        f"\n\n**Library name:** {package_name}"
        f"\n**{check_type.capitalize()} version:** {version}"
        f"\n**{check_type.capitalize()} Build:** [Link to build ({today.strftime('%Y-%m-%d')})]({build_link})"
        f"\n**How to fix:** Run the `next-{check_type}` azpysdk command at the library package-level and resolve "
        f"the {error_type} errors.\n"
        f"1) `{repo_root}>pip install -e ./eng/tools/azure-sdk-tools`\n"
        f"2) `../{package_name}>azpysdk next-{check_type} --isolate .`\n\n"
        f"See the {guide_link} for more information."
    )

    # create an issue for the library failing the vnext check
    if not vnext_issue:
        try:
            labels, assignees = get_labels(package_name, service_directory)
        except Exception as e:
            logging.warning(f"Failed to get labels and assignees from CODEOWNERS for {package_name}: {e}")
            labels = []
            assignees = []
            if "mgmt" in package_name:
                labels.append("Mgmt")

        labels.extend([check_type])
        logging.info(f"Issue does not exist for {package_name} with {check_type} version {version}. Creating...")
        issue = repo.create_issue(title=title, body=template, labels=labels)

        # Assign codeowners individually with error handling
        for assignee in assignees:
            try:
                issue.add_to_assignees(assignee)
                logging.info(f"Assigned {assignee} to issue for {package_name}")
            except GithubException as e:
                logging.warning(f"Failed to assign {assignee} to issue for {package_name}: {e}")
        return

    # an issue exists, let's update it so it reflects the latest typing/linting errors
    logging.info(f"Issue exists for {package_name} with {check_type} version {version}. Updating...")
    try:
        labels, assignees = get_labels(package_name, service_directory)
    except Exception as e:
        logging.warning(f"Failed to get labels and assignees from CODEOWNERS for {package_name}: {e}")
        labels = []
        assignees = []

    vnext_issue[0].edit(
        title=title,
        body=template,
    )

    # Assign codeowners individually with error handling
    for assignee in assignees:
        try:
            vnext_issue[0].add_to_assignees(assignee)
            logging.info(f"Assigned {assignee} to issue for {package_name}")
        except GithubException as e:
            logging.warning(f"Failed to assign {assignee} to issue for {package_name}: {e}")


def close_vnext_issue(package_name: str, check_type: CHECK_TYPE) -> None:
    """This is called when a client library passes a vnext check. If an issue exists for the library, it is closed."""

    auth = Auth.Token(os.environ["GH_TOKEN"])
    g = Github(auth=auth)

    repo = g.get_repo("Azure/azure-sdk-for-python")

    issues = repo.get_issues(state="open", labels=[check_type], creator="azure-sdk")
    vnext_issue = [issue for issue in issues if issue.title.split("needs")[0].strip() == package_name]
    if vnext_issue:
        logging.info(f"{package_name} passes {check_type}. Closing existing GH issue #{vnext_issue[0].number}...")
        vnext_issue[0].edit(state="closed")
